/*
 * $Id: template.c,v 1.4 2001/07/23 15:28:29 lindahl Exp $
 * 
 *                This source code is part of
 * 
 *                 G   R   O   M   A   C   S
 * 
 *          GROningen MAchine for Chemical Simulations
 * 
 *                        VERSION 3.0
 * 
 * Copyright (c) 1991-2001
 * BIOSON Research Institute, Dept. of Biophysical Chemistry
 * University of Groningen, The Netherlands
 * 
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 * 
 * If you want to redistribute modifications, please consider that
 * scientific software is very special. Version control is crucial -
 * bugs must be traceable. We will be happy to consider code for
 * inclusion in the official distribution, but derived work must not
 * be called official GROMACS. Details are found in the README & COPYING
 * files - if they are missing, get the official version at www.gromacs.org.
 * 
 * To help us fund GROMACS development, we humbly ask that you cite
 * the papers on the package - you can find them in the top README file.
 * 
 * Do check out http://www.gromacs.org , or mail us at gromacs@gromacs.org .
 * 
 * And Hey:
 * Gyas ROwers Mature At Cryogenic Speed
 */

/* This line is only for CVS version info */
static char *SRCID_template_c = "$Id: template.c,v 1.4 2001/07/23 15:28:29 lindahl Exp $";

#include "statutil.h"
#include "typedefs.h"
#include "smalloc.h"
#include "vec.h"
#include "copyrite.h"
#include "statutil.h"
#include "tpxio.h"
#include "string2.h"
#include "pbc.h"
#include "xvgr.h"
#include "math.h"
#include "physics.h" //for M_PI ...
#include "index.h"

/*#define __OLD_PBC */
/*#define _TEST_ */

#define grd(i,j,k) GRID[i][j][k]
#define dgrd(i,j,k) DGRID[i][j][k]
#ifdef __OLD_PBC
#define idgrd(i,j,k) (1-DGRID[(i+nbins[0])%nbins[0]][(j+nbins[1])%nbins[1]][(k+nbins[2])%nbins[2]])
#else
#define idgrd(i,j,k) (1-pbc_array(DGRID,i,j,k,nbins,offsets))
#endif

void getn3210(int*,ivec,int***,int**);

void freebin(int****, ivec);

void pbc_array_w(int**** array, int xxx, int yyy, int zzz, ivec nbins, int ** offsets, int value) {
//writes into array respecting PBC and possible offsets present in a triclinic system
  /*
  expected format of offsets:
  0  x-offset of b  x-offset of c
  0  0              y-offset of c 
  0  0              0
  */
  ivec coord;
  coord[0]=xxx;coord[1]=yyy;coord[2]=zzz;
  int i,j;
  for (i = 2 ; i >= 0 ; i--) {
    if (coord[i] >= nbins[i]) {
      coord[i]-=nbins[i];
      for (j = i-1 ; j >= 0 ; j--) {
        coord[j]-=offsets[i][j]; //+?
      }
    }
    else if (coord[i] < 0) {
      coord[i]+=nbins[i];
      for (j = i-1 ; j >= 0 ; j--) {
        coord[j]+=offsets[i][j]; //-?
      }
    }
  }
  (*array)[coord[0]][coord[1]][coord[2]]=value;
}

double dpbc_array(double*** array, int xxx, int yyy, int zzz, ivec nbins, int ** offsets) {
//adresses array respecting PBC and possible offsets present in a triclinic system
  /*
  expected format of offsets:
  0  x-offset of b  x-offset of c
  0  0              y-offset of c 
  0  0              0
  */
  ivec coord;
  coord[0]=xxx;coord[1]=yyy;coord[2]=zzz;
  int i,j;
  for (i = 2 ; i >= 0 ; i--) {
    if (coord[i] >= nbins[i]) {
      coord[i]-=nbins[i];
      for (j = i-1 ; j >= 0 ; j--) {
        coord[j]-=offsets[i][j]; //+?
      }
    }
    else if (coord[i] < 0) {
      coord[i]+=nbins[i];
      for (j = i-1 ; j >= 0 ; j--) {
        coord[j]+=offsets[i][j]; //-?
      }
    }
  }
  return array[coord[0]][coord[1]][coord[2]];
}

int pbc_array(int*** array, int xxx, int yyy, int zzz, ivec nbins, int ** offsets) {
//adresses array respecting PBC and possible offsets present in a triclinic system
  /*
  expected format of offsets:
  0  x-offset of b  x-offset of c
  0  0              y-offset of c 
  0  0              0
  */
  ivec coord;
  coord[0]=xxx;coord[1]=yyy;coord[2]=zzz;
  int i,j;
  for (i = 2 ; i >= 0 ; i--) {
    if (coord[i] >= nbins[i]) {
      coord[i]-=nbins[i];
      for (j = i-1 ; j >= 0 ; j--) {
        coord[j]-=offsets[i][j]; //+?
      }
    }
    else if (coord[i] < 0) {
      coord[i]+=nbins[i];
      for (j = i-1 ; j >= 0 ; j--) {
        coord[j]+=offsets[i][j]; //-?
      }
    }
  }
  return array[coord[0]][coord[1]][coord[2]];
}

void put_atoms_in_box_and_make_offset_multpl_dim(matrix box,int natoms,rvec x[],real dim)
{
  int i,m,d,j;
  matrix shift;
  for (i=1 ; i<=2 ; i++) {
    for (j=0 ; j<i ; j++) {
      shift[i][j] = box[i][j]-dim*(int)(box[i][j]/dim);
    }
  }

  for(i=0; (i<natoms); i++) {
    for(m=DIM-1; m>=0; m--) {
      while (x[i][m] < 0) 
	for(d=0; d<=m; d++)       //Starting with Z (m=2), the 2 whiles shift the atom by one boxvector box[m],
	  x[i][d] += box[m][d];   //changing the coordinates up to and including m
      while (x[i][m] >= box[m][m])
	for(d=0; d<=m; d++)
	  x[i][d] -= box[m][d];
      for (j=0 ; j<m ; j++) {
        x[i][j]-=shift[m][j]*x[i][m]/box[m][m]; //shifts lower coordinates for the current m (e.g. X and Y for Z)
      }                                         //doing this after the 2 whiles ensures, x[i][m] is in [0/box[m][m][
    }
  }
}

void write_cgo_frame( FILE *fp, real dx, real dy, real dz, real r )
{
    char *CGO_CYLINDER = "CYLINDER, %8.3f, %8.3f, %8.3f, %8.3f, %8.3f, %8.3f, %8.3f, 1, 1, 1, 1, 1, 1, \n";

    fprintf( fp, CGO_CYLINDER, 0.0, 0.0, 0.0,  dx, 0.0, 0.0, r );
    fprintf( fp, CGO_CYLINDER, 0.0, 0.0, 0.0, 0.0,  dy, 0.0, r );
    fprintf( fp, CGO_CYLINDER, 0.0, 0.0, 0.0, 0.0, 0.0,  dz, r );
    fprintf( fp, CGO_CYLINDER,  dx, 0.0, 0.0,  dx,  dy, 0.0, r );
    fprintf( fp, CGO_CYLINDER,  dx, 0.0, 0.0,  dx, 0.0,  dz, r );
    fprintf( fp, CGO_CYLINDER, 0.0,  dy, 0.0,  dx,  dy, 0.0, r );
    fprintf( fp, CGO_CYLINDER, 0.0,  dy, 0.0, 0.0,  dy,  dz, r );
    fprintf( fp, CGO_CYLINDER, 0.0, 0.0,  dz,  dx, 0.0,  dz, r );
    fprintf( fp, CGO_CYLINDER, 0.0, 0.0,  dz, 0.0,  dy,  dz, r );
    fprintf( fp, CGO_CYLINDER,  dx,  dy, 0.0,  dx,  dy,  dz, r );
    fprintf( fp, CGO_CYLINDER,  dx, 0.0,  dz,  dx,  dy,  dz, r );
    fprintf( fp, CGO_CYLINDER, 0.0,  dy,  dz,  dx,  dy,  dz, r );
}

void write_cgo_triangle( FILE *fp, real nx, real ny, real nz, real ax, real ay, real az, real bx, real by, real bz, real cx, real cy, real cz )
{
    fprintf( fp, "NORMAL, %8.3f, %8.3f, %8.3f, ", nx, ny, nz );
    fprintf( fp, "VERTEX, %8.3f, %8.3f, %8.3f, ", ax, ay, az );
    fprintf( fp, "VERTEX, %8.3f, %8.3f, %8.3f, ", bx, by, bz );
    fprintf( fp, "VERTEX, %8.3f, %8.3f, %8.3f, \n", cx, cy, cz );
}

void write_cgo_cube( FILE *fp, real x0, real y0, real z0, real d )
{
    real x1,y1,z1;

    x1 = x0 + d;
    y1 = y0 + d;
    z1 = z0 + d;

    fprintf( fp, "BEGIN, TRIANGLES, COLOR, 1, 1, 1, \n" );

    write_cgo_triangle( fp, 1, 0, 0, x1, y0, z0, x1, y0, z1, x1, y1, z0 );
    write_cgo_triangle( fp, 1, 0, 0, x1, y0, z1, x1, y1, z0, x1, y1, z1 );
    write_cgo_triangle( fp,-1, 0, 0, x0, y0, z0, x0, y0, z1, x0, y1, z0 );
    write_cgo_triangle( fp,-1, 0, 0, x0, y0, z1, x0, y1, z0, x0, y1, z1 );

    write_cgo_triangle( fp, 0, 1, 0, x0, y1, z0, x0, y1, z1, x1, y1, z0 );
    write_cgo_triangle( fp, 0, 1, 0, x0, y1, z1, x1, y1, z0, x1, y1, z1 );
    write_cgo_triangle( fp, 0,-1, 0, x0, y0, z0, x0, y0, z1, x1, y0, z0 );
    write_cgo_triangle( fp, 0,-1, 0, x0, y0, z1, x1, y0, z0, x1, y0, z1 );

    write_cgo_triangle( fp, 0, 0, 1, x0, y0, z1, x1, y0, z1, x0, y1, z1 );
    write_cgo_triangle( fp, 0, 0, 1, x0, y1, z1, x1, y0, z1, x1, y1, z1 );
    write_cgo_triangle( fp, 0, 0,-1, x0, y0, z0, x1, y0, z0, x0, y1, z0 );
    write_cgo_triangle( fp, 0, 0,-1, x0, y1, z0, x1, y0, z0, x1, y1, z0 );

    fprintf( fp, "END, \n" );
}
void write_color_cube( FILE *fp, real x0, real y0, real z0, real d, real color, /*for white 0 curv*/real curv)
{
    real x1,y1,z1;

    x1 = x0 + d;
    y1 = y0 + d;
    z1 = z0 + d;

    if (curv < 0) { //color [0.0/0.5[
      fprintf( fp, "BEGIN, TRIANGLES, COLOR, %3.2f, %3.2f, 1.00, \n",0.0+2*color,0.0+2*color );
      //if (color < 0 || color >= 0.5) { //MF DEBUG
      //  printf ("color range error: curv %14.12f color %3.2f \n",curv,color);
      //}
    } else { //color [0.5/1.0]
      fprintf( fp, "BEGIN, TRIANGLES, COLOR, 1.00, %3.2f, %3.2f, \n",1.0-2*(color-0.5),1.0-2*(color-0.5) );
      //if (color < 0.5 || color > 1.0) { //MF DEBUG
      //  printf ("color range error: curv %14.12f color %3.2f \n",curv,color);
      //}
    }

    write_cgo_triangle( fp, 1, 0, 0, x1, y0, z0, x1, y0, z1, x1, y1, z0 );
    write_cgo_triangle( fp, 1, 0, 0, x1, y0, z1, x1, y1, z0, x1, y1, z1 );
    write_cgo_triangle( fp,-1, 0, 0, x0, y0, z0, x0, y0, z1, x0, y1, z0 );
    write_cgo_triangle( fp,-1, 0, 0, x0, y0, z1, x0, y1, z0, x0, y1, z1 );

    write_cgo_triangle( fp, 0, 1, 0, x0, y1, z0, x0, y1, z1, x1, y1, z0 );
    write_cgo_triangle( fp, 0, 1, 0, x0, y1, z1, x1, y1, z0, x1, y1, z1 );
    write_cgo_triangle( fp, 0,-1, 0, x0, y0, z0, x0, y0, z1, x1, y0, z0 );
    write_cgo_triangle( fp, 0,-1, 0, x0, y0, z1, x1, y0, z0, x1, y0, z1 );

    write_cgo_triangle( fp, 0, 0, 1, x0, y0, z1, x1, y0, z1, x0, y1, z1 );
    write_cgo_triangle( fp, 0, 0, 1, x0, y1, z1, x1, y0, z1, x1, y1, z1 );
    write_cgo_triangle( fp, 0, 0,-1, x0, y0, z0, x1, y0, z0, x0, y1, z0 );
    write_cgo_triangle( fp, 0, 0,-1, x0, y1, z0, x1, y0, z0, x1, y1, z0 );

    fprintf( fp, "END, \n" );
}

void write_cgo_array( FILE *fp, int ***array, ivec bins, real d )
{
    int i,j,k;
    real dx,dy,dz;

    fprintf( fp, "from pymol.cgo import * \nfrom pymol import cmd \n\nobj = [ \n" );
    write_cgo_frame( fp, bins[0]*d, bins[1]*d, bins[2]*d, 0.1 );

    for ( i=0; i<bins[0]; i++ )
    {
	dx = i*d;
	for ( j=0; j<bins[1]; j++ )
	{
	    dy = j*d;
	    for ( k=0; k<bins[2]; k++ )
		if ( array[i][j][k] )
		{
		    dz = k*d;
		    write_cgo_cube( fp, dx, dy, dz, d );
		}
	}
    }
    
    fprintf( fp, "]\n\ncmd.load_cgo(obj, \"array3d\")\n" );
}
void write_color_image( FILE *fp, int ***image, ivec bins, real d, double *** curv, double scale, double min)
{
  int i,j,k;
  real dx,dy,dz;
  real color;

  fprintf( fp, "from pymol.cgo import * \nfrom pymol import cmd \n\nobj = [ \n" );
  write_cgo_frame( fp, bins[0]*d, bins[1]*d, bins[2]*d, 0.1 );

  for ( i=0; i<bins[0]; i++ )
  {
      dx = i*d;
      for ( j=0; j<bins[1]; j++ )
      {
          dy = j*d;
          for ( k=0; k<bins[2]; k++ )
      	if ( image[i][j][k] )
      	{
      	    dz = k*d;
            //compute color
            color=(curv[i][j][k]-min)*scale;
            if (color > 1.0) {
              color = 1.0;
              printf("WARNING (i) - color > 1.0: %18.16f\n",color);
            }
            else if (color < 0.0) {
              color = 0.0;
              printf("WARNING (i) - color < 0.0: %18.16f\n",color);
            }
      	    write_color_cube( fp, dx, dy, dz, d, color, curv[i][j][k]/*for white 0 curv*/);
      	}
      }
  }
  
  fprintf( fp, "]\n\ncmd.load_cgo(obj, \"array3d\")\n" );
}
void write_color_array( FILE *fp, double ***area, ivec bins, real d, double *** curv, double scale, double min)
{
    int i,j,k;
    real dx,dy,dz;
    real color;

    fprintf( fp, "from pymol.cgo import * \nfrom pymol import cmd \n\nobj = [ \n" );
    write_cgo_frame( fp, bins[0]*d, bins[1]*d, bins[2]*d, 0.1 );

    for ( i=0; i<bins[0]; i++ )
    {
	dx = i*d;
	for ( j=0; j<bins[1]; j++ )
	{
	    dy = j*d;
	    for ( k=0; k<bins[2]; k++ ) {
		if ( area[i][j][k] > 0.0 )
		{
		    dz = k*d;
                    //compute color
                    color=(curv[i][j][k]-min)*scale;
                    if (color > 1.0) {
                      color = 1.0;
                      printf("WARNING - color > 1.0: %18.16f\n",color);
                    }
                    else if (color < 0.0) {
                      color = 0.0;
                      printf("WARNING - color < 0.0: %18.16f\n",color);
                    }
		    write_color_cube( fp, dx, dy, dz, d, color, curv[i][j][k]/*for white 0 curv*/);
		}
            }
	}
    }
    
    fprintf( fp, "]\n\ncmd.load_cgo(obj, \"array3d\")\n" );
}

/*bool* makePhaseList(t_topology top,char ** negphase, int nres) {  
  bool* list;
  snew(list,top.atoms.nr);
  if (nres<1||nres>3) {
    printf("Error: Number of residues (%d) must be 1, 2 or 3.\n",nres);
    exit(1);
  }
  int atnum;
  for (atnum = 0 ; atnum < top.atoms.nr ; atnum++) { //smaller equal???
    list[atnum]=TRUE;
    int i;
    for (i = 0 ; i < nres ; i++) {
      //printf("resname %s negphase %s\n",*top.atoms.resname[top.atoms.atom[atnum].resnr],negphase[i]);
      if (strcmp(*top.atoms.resname[top.atoms.atom[atnum].resnr],negphase[i])==0) {
        //printf("WeGotPhase!\n");
        list[atnum]=FALSE;
      }
    }
    //printf("residue %s list %d resnr %d atnum %d\n",*top.atoms.resname[top.atoms.atom[atnum].resnr],list[atnum],top.atoms.atom[atnum].resnr,atnum);
  }
  return list;
}*/

/* radius sphere, number of points, POINT in grid(real space), box */
rvec *probe_spheres(real sigma, int *npoints,rvec POINT,matrix box, real dim)
{ 
  //  printf("*npoints3 %d\n",*npoints);
  int LOOP=0,i,j,k,next=0;
  static rvec *Xp = NULL;
  static real dSIGMA =0, dTHETA=0, dPHI = 0;
  real r=0,theta=0, phi=0;

  rvec *Xtemp;
 
  real x,y,z;
  int ax1,ax2,ax3;

  x = POINT[XX];
  y = POINT[YY];
  z = POINT[ZZ];

  ax1=XX;
  ax2=YY;
  ax3=ZZ;

  snew(Xtemp,*npoints);


  /* this is only done the first time*/

  if(*Xp == NULL)  //changed Xp to *Xp should be fine, but MF CHECK anyway! MEMORY
  { 
    printf("Xp == NULL\n"); //MF

 
    /*cubic root*/
    LOOP =(int)(pow(*npoints,(1.0/3.0))+0.1);

    /* redefine*/

    *npoints = LOOP*LOOP*LOOP;
    printf("*npoints4 %d\n",*npoints);

    snew(Xp,LOOP*LOOP*LOOP);

    dSIGMA = (real)sigma / LOOP;
    dTHETA = (real)(M_PI)/ LOOP;
    dPHI   = (real)(M_PI*2.0) / LOOP;

    next =0;
    for(i=0;i<LOOP;i++) //r-loop
    {
      phi = 0; //MF
      for(j=0;j<LOOP;j++) //phi-loop
      {
        theta = 0; //MF
        for(k=0;k<LOOP;k++) //theta-loop
        {

          Xp[next][ax1]= r*cos(theta)*sin(phi);
          Xp[next][ax2]= r*sin(theta)*sin(phi);
          Xp[next][ax3]= r*cos(phi);

          theta += dTHETA;
          next++;
        }
 
        phi += dPHI;

      }
      r += dSIGMA;

    }
  }

  /* now translate the reference sphere to the position of the atom*/
  /*put everyting in the box periodically !!!*/

  /*THIS IS THE ACTUAL MAKING OF THE SPHERE POINTS FOR THE INPUT COORDINATE*/
  for(i=0;i<*npoints;i++)
  {

    Xtemp[i][ax1] = Xp[i][ax1] + x;
//    while(Xtemp[i][ax1] >= box[ax1][ax1]) //changed > to >= (MF)
//      Xtemp[i][ax1] -= box[ax1][ax1];
//    while(Xtemp[i][ax1] < 0)
//      Xtemp[i][ax1] += box[ax1][ax1];
  
    Xtemp[i][ax2] = Xp[i][ax2] + y;
//    while(Xtemp[i][ax2] >= box[ax2][ax2]) //changed > to >= (MF)
//      Xtemp[i][ax2] -= box[ax2][ax2];
//    while(Xtemp[i][ax2] < 0)
//      Xtemp[i][ax2] += box[ax2][ax2];
 
    Xtemp[i][ax3] = Xp[i][ax3] + z;
//    while(Xtemp[i][ax3] >= box[ax3][ax3]) //changed > to >= (MF)
//      Xtemp[i][ax3] -= box[ax3][ax3];
//    while(Xtemp[i][ax3] < 0)
//      Xtemp[i][ax3] += box[ax3][ax3];

  }
  //put_atoms_in_box(box,*npoints,Xtemp); //MF CHECK
  put_atoms_in_box_and_make_offset_multpl_dim(box,*npoints,Xtemp,dim); //MF CHECK


  /* now do the grid indexing over this array of *npoints*/
  return Xtemp;
}

int*** get_locID(int*** bin ,ivec nbins,int nlc,double chi, int** offsets, real* global_curv) {
  int nn,phase,i,j,k,plusminus,s,t,u;
  double chi2;
  double maxMean, minMean, maxGauss, minGauss;
  chi2 = chi*chi;
  int*     oct3210; //number of cubes, faces, edges, vertices used to identify octant for local Curvatures
  snew(oct3210,4);
  plusminus = 1;
  ivec tmp;
  ivec noct;
  ivec nbinsnlc;
  int** zeromatrix;
  snew(zeromatrix,3);
  for (i=0 ; i<3 ; i++) {
    printf("nbins[%d]=%d\n",i,nbins[i]);
    for (j=0;j<3;j++) {
      printf("offsets[%d][%d]=%d\n",i,j,offsets[i][j]);
    }
    noct[i]=3; 
    nbinsnlc[i]=nbins[i]/nlc;
    snew(zeromatrix[i],3);
    for (j=0 ; j<3 ; j++) {
      zeromatrix[i][j]=0;
    }
  }
  //allocate bin
  int*** mybin;
  snew(mybin,(nbins)[0]);     
  for (i = 0 ; i<(nbins)[0] ; i++) {
    snew(mybin[i],(nbins)[1]);
    for (j = 0 ; j<(nbins)[1] ; j++) {
      snew(mybin[i][j],(nbins)[2]); 
      /*TEST*/
      ///*
#ifdef  _TEST_
      for (k=0;k<nbins[2];k++) {
        if (FALSE /*j==5 || j==6 || j==4 || j==7*/) {
          bin[i][j][k]=1;
        } else {
          bin[i][j][k]=0;
        }
      }
      //*/
#endif
      /*TEST*/
    }
  }
  //TEST
#ifdef _TEST_
  ///*
  //bin[0][0][0]=1;
  //bin[0][1][2]=1;
  //bin[4][nbins[1]-1][4]=1;
  //bin[4][nbins[1]-2][4]=1;
  //bin[4][nbins[1]-3][4]=1;
  //bin[4+offsets[1][0]][0+0][4]=1;
  //bin[4+offsets[1][0]][0+1][4]=1;
  //bin[4+offsets[1][0]][0+2][4]=1;
  for(i=0;i<10;i++) {
    for(j=0;j<10;j++) {
      for(k=0;k<10;k++) {
      bin[4+i][4+j][nbins[2]-k-1]=1;
      //bin[4+i][4+j][nbins[2]-2]=1;
      //bin[4+i][4+j][nbins[2]-3]=1;
      bin[4+i+offsets[2][0]][4+j+offsets[2][1]][0+k]=1;
      //bin[4+i+offsets[2][0]][4+j+offsets[2][1]][0+1]=1;
      //bin[4+i+offsets[2][0]][4+j+offsets[2][1]][0+2]=1;
      }
    }
  }
  //*/
  /*
  bin[4][3][4]=1;
  bin[5][3][4]=1;
  bin[4][3][5]=1;
  bin[5][3][5]=1;
  bin[1][4][1]=0;
  bin[2][4][1]=0;
  bin[1][4][2]=0;
  bin[2][4][2]=0;
  bin[0][9][6]=1;
  bin[20][5][20]=0;
  bin[20][6][20]=0;
  bin[20][4][20]=0;
  bin[20][7][20]=0;
  bin[25][5][20]=0;
  bin[25][6][20]=0;
  bin[25][4][20]=0;
  bin[25][7][20]=0;
  bin[25][5][21]=0;
  bin[25][6][21]=0;
  bin[25][4][21]=0;
  bin[25][7][21]=0;
  bin[26][5][20]=0;
  bin[26][6][20]=0;
  bin[26][4][20]=0;
  bin[26][7][20]=0;
  bin[26][5][21]=0;
  bin[26][6][21]=0;
  bin[26][4][21]=0;
  bin[26][7][21]=0;

  bin[15][21][20]=1;
  bin[15][22][20]=1;
  bin[15][23][20]=1;
  bin[15][24][20]=1;
  bin[15][25][20]=1;
  bin[15][20][20]=1;

  bin[15][15][21]=1;
  bin[15][15][22]=1;
  bin[15][15][23]=1;
  bin[15][15][24]=1;
  bin[15][15][25]=1;
  bin[15][15][20]=1;

  bin[20][20][20]=1;
  bin[21][20][20]=1;
  bin[22][20][20]=1;
  bin[23][20][20]=1;
  bin[24][20][20]=1;
  bin[25][20][20]=1;

  bin[1][1][1]=1;
  */
#endif
  //TEST
  double*** meanbin,***gaussbin,***areabin;
  snew(gaussbin,nbins[0]);     // it has already been assured that nbins is a multiple of nlc...
  snew(meanbin,nbins[0]);
  snew(areabin,nbins[0]);
  for (i = 0 ; i<nbins[0] ; i++) {
    snew(gaussbin[i],nbins[1]);
    snew(meanbin[i],nbins[1]);
    snew(areabin[i],nbins[1]);
    for (j = 0 ; j<nbins[1] ; j++) {
      snew(gaussbin[i][j],nbins[2]); 
      snew(meanbin[i][j],nbins[2]); 
      snew(areabin[i][j],nbins[2]); 
    }
  }
  int *** tmpoctant;
  snew(tmpoctant,noct[0]);
  for (i = 0 ; i<noct[0] ; i++) { 
    snew(tmpoctant[i],noct[1]);
    for (j = 0 ; j<noct[1] ; j++) {
      snew(tmpoctant[i][j],noct[2]); 
    }
  }
  for (i = 0 ; i<noct[0] ; i++) {
    for (j = 0 ; j<noct[1] ; j++) {
      for (k = 0 ; k<noct[2] ; k++) {
        tmpoctant[i][j][k]=0;
      }
    }
  }
#ifdef __OLD_PBC
#define binpbc(s,t,u) bin[(s+nbins[0])%nbins[0]][(t+nbins[1])%nbins[1]][(u+nbins[2])%nbins[2]]
#define newbinpbc(s,t,u) newbin[(s+nbins[0])%nbins[0]][(t+nbins[1])%nbins[1]][(u+nbins[2])%nbins[2]]
#define meanbinpbc(s,t,u) meanbin[(s+nbins[0])%nbins[0]][(t+nbins[1])%nbins[1]][(u+nbins[2])%nbins[2]]
#define areabinpbc(s,t,u) areabin[(s+nbins[0])%nbins[0]][(t+nbins[1])%nbins[1]][(u+nbins[2])%nbins[2]]
#define ColIpbc(s,t,u) ColImage[(s+nbins[0])%nbins[0]][(t+nbins[1])%nbins[1]][(u+nbins[2])%nbins[2]]
#else
#define binpbc(s,t,u) pbc_array(bin,s,t,u,nbins,offsets)
#define newbinpbc(s,t,u) pbc_array(newbin,s,t,u,nbins,offsets)
#define meanbinpbc(s,t,u) dpbc_array(meanbin,s,t,u,nbins,offsets)
#define gaussbinpbc(s,t,u) dpbc_array(gaussbin,s,t,u,nbins,offsets)
#define areabinpbc(s,t,u) dpbc_array(areabin,s,t,u,nbins,offsets)
#define ColIpbc(s,t,u) dpbc_array(ColImage,s,t,u,nbins,offsets)
#define gColIpbc(s,t,u) dpbc_array(gColImage,s,t,u,nbins,offsets)
#endif
#define GSSBN(s,t,u) gaussbin[/*(int)(*/s/*/nlc)*/][/*(int)(*/t/*/nlc)*/][/*(int)(*/u/*/nlc)*/]
#define MNBN(s,t,u)   meanbin[/*(int)(*/s/*/nlc)*/][/*(int)(*/t/*/nlc)*/][/*(int)(*/u/*/nlc)*/]
#define RBN(s,t,u)    areabin[/*(int)(*/s/*/nlc)*/][/*(int)(*/t/*/nlc)*/][/*(int)(*/u/*/nlc)*/]
  //loop over bin and setup tmpoctant...
  for (i = 0 ; i < nbins[0] ; i++) {
    for (j = 0 ; j < nbins[1] ; j++) {
      for (k = 0 ; k < nbins[2] ; k++) {
        for (s = 0 ; s < 2 ; s++) { //first set up tmpoctant for every bin...
          for (t = 0 ; t < 2 ; t++) {
            for (u = 0 ; u < 2 ; u++) {
              tmpoctant[s][t][u] = binpbc(i+s,j+t,k+u);
            }
          }
        }
        //then identify tmpoctant
        //MF  here also offsets appeared now = zeromatrix...
        getn3210(oct3210,noct,tmpoctant,zeromatrix); //just a handy way of discriminating, the actual numbers have no meaning
        mybin[i][j][k]=oct3210[3]+oct3210[2]+oct3210[1]+oct3210[0]; //maybe this is not ideal...
        if (mybin[i][j][k]==27) { //case a
          GSSBN(i,j,k)+= 0.5*M_PI; 
          MNBN(i,j,k)+= M_PI*0.375*chi; //THE CURVATURES ARE ALREADY MULTIPLIED WITH THE AREA!!
          RBN(i,j,k)+= 0.75*chi2;
          global_curv[3]+= M_PI*0.375*chi;
          global_curv[4]+= 0.5*M_PI;
          global_curv[2]+= 0.75*chi2;
          printf("CASE a\n");
        }
        else if (mybin[i][j][k]==117) { //case ai
          GSSBN(i,j,k)+= 0.5*M_PI;
          MNBN(i,j,k)+= -M_PI*0.375*chi;
          RBN(i,j,k)+= 0.75*chi2;
          global_curv[3]+=-M_PI*0.375*chi;
          global_curv[4]+=0.5*M_PI;
          global_curv[2]+= 0.75*chi2;
          printf("CASE ai\n");
        }
        else if (mybin[i][j][k]==45) { //case b
          //GSSBN(i,j,k)+= 0;
          MNBN(i,j,k)+= 0.25*chi*M_PI;
          RBN(i,j,k)+= chi2;
          global_curv[3]+=0.25*chi*M_PI;
          //global_curv[4]+=
          global_curv[2]+= chi2;
          printf("CASE b\n");
        }
        else if (mybin[i][j][k]==105) { //case bi
          //GSSBN(i,j,k)+= 0;
          MNBN(i,j,k)+= -0.25*chi*M_PI;
          RBN(i,j,k)+= chi2;
          global_curv[3]+= -0.25*chi*M_PI;
          //global_curv[4]+=
          global_curv[2]+= chi2;
          printf("CASE bi\n");
        }
        else if (mybin[i][j][k]==51) { //case c
          GSSBN(i,j,k)+= -M_PI;
          MNBN(i,j,k)+= 0.25*chi*M_PI;
          RBN(i,j,k)+= 1.5*chi2;
          global_curv[3]+= 0.25*chi*M_PI;
          global_curv[4]+= -M_PI;
          global_curv[2]+= 1.5*chi2;
          printf("CASE c\n");
        }
        else if (mybin[i][j][k]==109) { //case ci
          GSSBN(i,j,k)+= M_PI;
          MNBN(i,j,k)+= -0.75*chi*M_PI;
          //GSSBN(i,j,k)+= -M_PI;
          //MNBN(i,j,k)+= -0.25*chi*M_PI;
          RBN(i,j,k)+= 1.5*chi2;
          global_curv[3]+=-0.75*chi*M_PI;
          global_curv[4]+= M_PI;
          global_curv[2]+= 1.5*chi2;
          printf("CASE ci\n");
        }
        else if (mybin[i][j][k]==63) { //case e
          GSSBN(i,j,k)+= -0.5*M_PI;
          MNBN(i,j,k)+= 0.125*chi*M_PI;
          RBN(i,j,k)+= 1.25*chi2;
          global_curv[3]+=0.125*chi*M_PI;
          global_curv[4]+=-0.5*M_PI;
          global_curv[2]+= 1.25*chi2;
          printf("CASE e\n");
        }
        else if (mybin[i][j][k]==93) { //case ei
          GSSBN(i,j,k)+= -0.5*M_PI;
          MNBN(i,j,k)+= -0.125*chi*M_PI;
          RBN(i,j,k)+= 1.25*chi2;
          global_curv[3]+=-0.125*chi*M_PI;
          global_curv[4]+=-0.5*M_PI;
          global_curv[2]+= 1.25*chi2;
          printf("CASE ei\n");
        }
        else if (mybin[i][j][k]==69) { //case f
          GSSBN(i,j,k)+= -1.5*M_PI;
          MNBN(i,j,k)+= 0.125*chi*M_PI;
          RBN(i,j,k)+= 1.75*chi2;
          global_curv[3]+=0.125*chi*M_PI;
          global_curv[4]+=-1.5*M_PI;
          global_curv[2]+= 1.75*chi2;
          printf("CASE f\n");
        }
        else if (mybin[i][j][k]==97) { //case fi
          GSSBN(i,j,k)+= 0.5*M_PI;
          MNBN(i,j,k)+= -0.625*chi*M_PI;
          //GSSBN(i,j,k)+= -1.5*M_PI;
          //MNBN(i,j,k)+= -0.125*chi*M_PI;
          RBN(i,j,k)+= 1.75*chi2;
          global_curv[3]+=-0.625*chi*M_PI;
          global_curv[4]+=0.5*M_PI;
          global_curv[2]+= 1.75*chi2;
          printf("CASE fi\n");
        }
        else if (mybin[i][j][k]==75) { //case h
          //GSSBN(i,j,k)+= 0;
          //MNBN(i,j,k)+= 0;
          RBN(i,j,k)+= chi2;
          //global_curv[3]+=
          //global_curv[4]+=
          global_curv[2]+= chi2;
          printf("CASE h\n");
        }
        else if (mybin[i][j][k]==81) { //case i
          GSSBN(i,j,k)+= -M_PI;
          //MNBN(i,j,k)+= 0;
          RBN(i,j,k)+= 1.5*chi2;
          //global_curv[3]+=
          global_curv[4]+=-M_PI;
          global_curv[2]+= 1.5*chi2;
          printf("CASE i\n");
        }
        else if (mybin[i][j][k]!=0 && mybin[i][j][k]!=125) {
          printf("CASE NOISE %d\n",mybin[i][j][k]);
        }
      }
    }
  }
  double scale;
  minMean=-3*M_PI*chi/*EXPERIMENTAL*//8; //MF CHECK THIS! for now leave it for loc-curv.py (event. delete it)
  scale = -1.0/(2*minMean); //make scaling absolute, normalize by largest possible values, not the actual ones
  printf("6PIchi=%18.16f ,scale %18.16f\n",M_PI*6*chi,scale); //MF INFO OUTPUT
  char* fn3 = "loc-curv-vertices.py"; //attention! so far only works for a single frame!
  FILE* fp3;
  fp3 = fopen(fn3,"w");


  //write_color_array(fp3,areabin,nbinsnlc,chi*nlc/*dimension*/,meanbin,scale,minMean);
  write_color_array(fp3,areabin,nbins,chi/**nlc*//*dimension*/,meanbin,scale,minMean);

  //PREPROCESS IMAGE = DELETE NON-SURFACE PIXELS
  int *** newbin;
  snew(newbin,nbins[0]);
  for (i = 0 ; i<nbins[0] ; i++) {
    snew(newbin[i],nbins[1]);
    for (j = 0 ; j<nbins[1] ; j++) {
      snew(newbin[i][j],nbins[2]); 
    }
  }
  for ( i=0; i<nbins[0]; i++ ) {
    for ( j=0; j<nbins[1]; j++ ) {
      for ( k=0; k<nbins[2]; k++ ) {
        if (bin[i][j][k]) {
          if (!(binpbc(i+1,j,k)&&binpbc(i-1,j,k)&&binpbc(i,j+1,k)&&binpbc(i,j-1,k)&&binpbc(i,j,k+1)&&binpbc(i,j,k-1))){
            newbin[i][j][k]=bin[i][j][k];
          }
        }
      }
    }
  }
      char* fn2 = "visual.py"; //attention! so far only works for a single frame!
      FILE* fp2;
      fp2 = fopen(fn2,"w");
      write_cgo_array(fp2,newbin,nbins,chi);
      fclose(fp2);

  //MAP CURVATURE ON PIXELS
  double *** ColImage, *** nColImage, *** gColImage, *** ngColImage;
  snew(ColImage,nbins[0]);
  snew(nColImage,nbins[0]);
  snew(gColImage,nbins[0]);
  snew(ngColImage,nbins[0]);
  for (i = 0 ; i<nbins[0] ; i++) {
    snew(ColImage[i],nbins[1]);
    snew(nColImage[i],nbins[1]);
    snew(gColImage[i],nbins[1]);
    snew(ngColImage[i],nbins[1]);
    for (j = 0 ; j<nbins[1] ; j++) {
      snew(ColImage[i][j],nbins[2]); 
      snew(nColImage[i][j],nbins[2]); 
      snew(gColImage[i][j],nbins[2]); 
      snew(ngColImage[i][j],nbins[2]); 
    }
  }
  //loop over areabin = VERTICES ----> count adjacent surface PIXELS
  //store in areabin
  for (i=0 ; i<nbins[0] ; i++) {
    for (j=0 ; j<nbins[1] ; j++) {
      for (k=0 ; k<nbins[2] ; k++) {
        if(areabin[i][j][k]>0.0) {
          areabin[i][j][k]=0.2; //safety
          //printf("counting surf pix i %d j %d k %d\n",i,j,k); //MFDEBUG
          for (s=0 ; s<=1 ; s++) {
            for (t=0 ; t<=1 ; t++) {
              for (u=0 ; u<=1 ; u++) {
                if (newbinpbc(i+s,j+t,k+u)) {
                  areabin[i][j][k]+=1.0; // areabin is still over VERTICES - contains # adjacent surface pixels
                  //printf("areabin[%d][%d][%d]=%16.14f\n",i,j,k,areabin[i][j][k]); //MFDEBUG
                }
              }
            }
          }
        }
      }
    }
  }
  //loop over PIXELS
  minMean=0.0; maxMean=0.0;
  maxGauss=0.0; minGauss=0.0;
  for (i=0 ; i<nbins[0] ; i++) {
    for (j=0 ; j<nbins[1] ; j++) {
      for (k=0 ; k<nbins[2] ; k++) {
        if(newbin[i][j][k]) {
          for (s=-1 ; s<=0 ; s++) { //loop over adjacent vertices
            for (t=-1 ; t<=0 ; t++) {
              for (u=-1 ; u<=0 ; u++) {
                if (areabinpbc(i+s,j+t,k+u)>0.3) {//=check if this is a surface vertice
                  ColImage[i][j][k]+=meanbinpbc(i+s,j+t,k+u)/(int)(areabinpbc(i+s,j+t,k+u)); //distribute curv of vertices
                  gColImage[i][j][k]+=gaussbinpbc(i+s,j+t,k+u)/(int)(areabinpbc(i+s,j+t,k+u)); //distribute curv of vertices
                  global_curv[0]+=meanbinpbc(i+s,j+t,k+u)/(int)(areabinpbc(i+s,j+t,k+u));
                  global_curv[1]+=gaussbinpbc(i+s,j+t,k+u)/(int)(areabinpbc(i+s,j+t,k+u));
                  //printf("CI %10.9f mb %10.9f ab %d\ni:%d j:%d k:%d, s:%d t:%d u:%d\n",ColImage[i][j][k],meanbinpbc(i+s,j+t,k+u),(int)(areabinpbc(i+s,j+t,k+u)),i,j,k,s,t,u);//MF DEBUG
                }                                             //in equal parts on adjacent pixels (areabin = n.o.pixels,
              }                                              //not area!)
            }
          }
          //get max/min value:
          if (ColImage[i][j][k]>maxMean) maxMean=ColImage[i][j][k];
          if (ColImage[i][j][k]<minMean) minMean=ColImage[i][j][k]; 
          if (gColImage[i][j][k]>maxGauss) maxGauss=gColImage[i][j][k];
          if (gColImage[i][j][k]<minGauss) minGauss=gColImage[i][j][k]; 
        }
      }
    }
  }

  char* fn4 = "Mean-Curv.py"; //attention! so far only works for a single frame!
  FILE* fp4;
  fp4 = fopen(fn4,"w");
  char* fn5 = "Gauss-Curv.py"; //attention! so far only works for a single frame!
  FILE* fp5;
  fp5 = fopen(fn5,"w");

  //IF YOU WANT AVERAGING, DO IT NOW...
  int counter;
  if (nlc>=1) { //MF LAST CHANGE
    //minMean=3*M_PI*chi; maxMean=-minMean; //reset min/max
    minMean=0.0; maxMean=0.0;
    maxGauss=0.0; minGauss=0.0;
    for (i=0 ; i<nbins[0] ; i++) {
      for (j=0 ; j<nbins[1] ; j++) {
        for (k=0 ; k<nbins[2] ; k++) {
          if (newbin[i][j][k]) {
            counter=0;
            for (s=-nlc ; s<=nlc ; s++) {
              for (t=-nlc ; t<=nlc ; t++) {
                for (u=-nlc ; u<=nlc ; u++) {
                  if (newbinpbc(i+s,j+t,k+u)) {
                    counter++;
                  }
                }
              }
            }
          //NColi/=counter
          ColImage[i][j][k]/=counter;
          gColImage[i][j][k]/=counter;
          }
        }
      }
    }
    for (i=0 ; i<nbins[0] ; i++) {
      for (j=0 ; j<nbins[1] ; j++) {
        for (k=0 ; k<nbins[2] ; k++) {
          if (newbin[i][j][k]) {
            for (s=-nlc ; s<=nlc ; s++) {
              for (t=-nlc ; t<=nlc ; t++) {
                for (u=-nlc ; u<=nlc ; u++) {
                  if (newbinpbc(i+s,j+t,k+u)) {
                    nColImage[i][j][k]+=ColIpbc(i+s,j+t,k+u);
                    ngColImage[i][j][k]+=gColIpbc(i+s,j+t,k+u);
                  }
                }
              }
            }
            if (nColImage[i][j][k]>maxMean) maxMean=nColImage[i][j][k];
            if (nColImage[i][j][k]<minMean) minMean=nColImage[i][j][k]; // no else...
            if (ngColImage[i][j][k]>maxGauss) maxGauss=ngColImage[i][j][k];
            if (ngColImage[i][j][k]<minGauss) minGauss=ngColImage[i][j][k]; // no else...
          }
        }
      }
    }
  }
  //MEAN CURV PIC
  if (minMean==0.0 && maxMean==0) { //no div by 0...
    scale = 1.0;
  } else {
    printf("minMean %18.17f  maxMean %18.17f\n",minMean,maxMean);//MF DEBUG
    if (fabs(minMean)>fabs(maxMean)) {
      scale = fabs(1.0/(2*minMean));
      minMean = - fabs(minMean);
    } else {
      scale = fabs(1.0/(2*maxMean));
      minMean = - fabs(maxMean);
    }
    printf("minMean %18.17f  maxMean %18.17f scale %18.17f\n",minMean,maxMean,scale);//MF DEBUG
  }

  if (nlc>=1) { //MF LAST CHANGE
    write_color_image(fp4,newbin,nbins,chi/*dimension*/,nColImage,scale,minMean);
  } else {
    write_color_image(fp4,newbin,nbins,chi/*dimension*/,ColImage,scale,minMean);
  }

  //GAUSS CURV PIC
  if (minGauss==0.0 && maxGauss==0) { //no div by 0...
    scale = 1.0;
  } else {
    printf("minGauss %18.17f  maxGauss %18.17f\n",minGauss,maxGauss);//MF DEBUG
    if (fabs(minGauss)>fabs(maxGauss)) {
      scale = fabs(1.0/(2*minGauss));
      minGauss = - fabs(minGauss);
    } else {
      scale = fabs(1.0/(2*maxGauss));
      minGauss = - fabs(maxGauss);
    }
    printf("minGauss %18.17f  maxGauss %18.17f scale %18.17f\n",minGauss,maxGauss,scale);//MF DEBUG
  }

  if (nlc>=1) { //MF LAST CHANGE
    write_color_image(fp5,newbin,nbins,chi/*dimension*/,ngColImage,scale,minGauss);
  } else {
    write_color_image(fp5,newbin,nbins,chi/*dimension*/,gColImage,scale,minGauss);
  }


  fclose(fp4);
  fclose(fp5);

  fclose(fp3);
  freebin(&tmpoctant,noct);
  freebin(&ColImage,nbins); //not completely nice - freebin expects a pointer to integers, bus sfree doesn't mind?
  freebin(&nColImage,nbins); //not completely nice - freebin expects a pointer to integers, bus sfree doesn't mind?
  freebin(&gColImage,nbins); //not completely nice - freebin expects a pointer to integers, bus sfree doesn't mind?
  freebin(&ngColImage,nbins); //not completely nice - freebin expects a pointer to integers, bus sfree doesn't mind?
  freebin(&newbin,nbins);
  sfree(oct3210);
  for (i = 0 ; i < 3 ; i++)
    sfree(zeromatrix[i]);
  sfree(zeromatrix);
  return mybin;
}

int*** binning(t_trxframe frame,atom_id **index,int*nindex,int np,real dim,ivec* nbins,int *npts,real sigma,ivec** ndx,int nlc, bool LV, int thresh, int cs, int*** offsets) {
  if (np<1||np>2) {
    printf("Error: np has to be 1 or 2!");
    exit(9);
  }
  int temp,nn,phase,i,j,k,plusminus;
  plusminus = 1;
  ivec tmp;
  for (i = 0 ; i<3 ; i++) { //MFDEBUG
    for (j = 0 ; j<3 ; j++) {
      printf("box[%d][%d]=%8.5f\n",i,j,frame.box[i][j]);
    }
  } //MFDEBUG
  snew(*offsets,3);
  for (i = 0 ; i<3 ; i++) {
    snew((*offsets)[i],3);
  }
  //determine number of bins
  for (i=0 ; i<DIM ; i++) {
    (*nbins)[i]=(int)(frame.box[i][i]/dim);
    if ((*nbins)[i]*dim>frame.box[i][i]) { //MFDEBUG took out = 
      printf("Error: (*nbins)[%d]*dim=%8.5f > frame.box[%d]=%8.5f\n",i,(*nbins)[i]*dim,i,frame.box[i][i]);
    }
    fprintf(stderr,"(*nbins) %d i %d natoms %d\n",(*nbins)[i],i,frame.natoms); //MF DEBUG
  }
  //allocate bin
  int*** mybin;
  rvec* sphere;
  snew(mybin,(*nbins)[0]);
  for (i = 0 ; i<(*nbins)[0] ; i++) {
    snew(mybin[i],(*nbins)[1]);
    for (j = 0 ; j<(*nbins)[1] ; j++) {
      snew(mybin[i][j],(*nbins)[2]); 
    }
  }
  //determine offsets
  for (i=0 ; i<3 ; i++) {
    for (j=0 ; j<i ; j++) {
      (*offsets)[i][j] = (int)(frame.box[i][j]/dim);
    }
  }

  //put atoms in box (might want to check if it works for triclinic boxes...)
  //put_atoms_in_box(frame.box,frame.natoms,frame.x);
  put_atoms_in_box_and_make_offset_multpl_dim(frame.box,frame.natoms,frame.x,dim);

  //loop over atoms in index
  for (phase = 0 ; phase < np ; phase++) {  
    for (nn=0 ; nn < nindex[phase] ; nn++) {
      //expand atom to sphere (not one coordinate, but npts points)
      sphere = probe_spheres(sigma, npts, frame.x[index[phase][nn]],frame.box,dim);
      for (j = 0 ; j < *npts ; j++) {
        for (i = 0 ; i < DIM ; i++) {
          tmp[i]=(int)(sphere[j][i]/frame.box[i][i]*(*nbins)[i]);
          if (tmp[i]==(*nbins)[i]) { //MF SAFETY MEASURE
            fprintf(stderr,"tmp: %d, box: %d\nreducing tmp by 1!\n",tmp[i],(*nbins)[i]);
            tmp[i]--;
          }
        }
        mybin[tmp[0]][tmp[1]][tmp[2]]+=plusminus;
      }
      sfree(sphere);
    }
    plusminus=-1; //we do the above once for the positive phase. if the negative phase is treated explicitely
		  //(np = 2) we do this ONE more time, this time subtracting...
  }
  //convert bin to 0 or 1 - format expected in getn3210
  //and create ndx of background... which is needed for the cluster filtering
  snew(*ndx,(*nbins)[0]*(*nbins)[1]*(*nbins)[2]+1);//ndx: 0 length([0][0]), 1,2,3... indices of nth element
  (*ndx)[0][0]=1;
  
  for (i = 0 ; i<(*nbins)[0] ; i++) {
    for (j = 0 ; j<(*nbins)[1] ; j++) {
      for (k = 0 ; k<(*nbins)[2] ; k++) {
        if (mybin[i][j][k]>thresh) { //FOREGROUND
          mybin[i][j][k]=1;
          //DO I ALSO NEED AN NDX OF THE FOREGROUND?? MF IMAGEMANIPULATION...
        } else {                     //BACKGROUND
          if (cs>0) {
            (*ndx)[(*ndx)[0][0]][0]=i;
            (*ndx)[(*ndx)[0][0]][1]=j;
            (*ndx)[(*ndx)[0][0]][2]=k;
            mybin[i][j][k]=-(*ndx)[0][0]+1; //+1 to number the background from [0-#bgd[
            (*ndx)[0][0]++;
          }
          else {
            mybin[i][j][k]=0;
          }
        }
      }
    }
  }
  return mybin;
}

void freebin(int**** bin, ivec nbins) {
  int i,j;
  for (i = 0 ; i<nbins[0] ; i++) {
    for (j = 0 ; j<nbins[1] ; j++) {
      sfree((*bin)[i][j]); 
    }
    sfree((*bin)[i]);
  }
  sfree(*bin);
}

void getn3210(int* n3210,ivec nbins,int*** GRID,int** offsets) {//computes number of active cubes, faces, edges and vertices
                                                   
  int xx,yy,zz,aa,bb,cc,i,j;
  //allocate dual Grid
  int *** DGRID;
  snew(DGRID,nbins[0]);
  for (i = 0 ; i<nbins[0] ; i++) {
    snew(DGRID[i],nbins[1]);
    for (j = 0 ; j<nbins[1] ; j++) {
      snew(DGRID[i][j],nbins[2]); 
    }
  }
  //set n3210 to zero
  for (i = 0 ; i < 4 ; i++)
    n3210[i] = 0;
  //loop over Grid (uses definitions! idgrd contains modulus --> PBC)
  for (xx = 0 ; xx < nbins[0] ; xx++) {
    for (yy = 0 ; yy < nbins[1] ; yy++) {
      for (zz = 0 ; zz < nbins[2] ; zz++) {
        if (grd(xx,yy,zz)==1) {
          dgrd(xx,yy,zz)=1;
          n3210[3] += 1;
          for (aa = -1 ; aa <= 1 ; aa += 2) {
            n3210[2] += idgrd(xx+aa,yy,zz) + idgrd(xx,yy+aa,zz) + idgrd(xx,yy,zz+aa);
            for (bb = -1 ; bb <= 1 ; bb += 2) {
              n3210[1] += idgrd(xx+aa,yy,zz)*idgrd(xx+aa,yy+bb,zz)*idgrd(xx,yy+bb,zz)
                        + idgrd(xx+aa,yy,zz)*idgrd(xx+aa,yy,zz+bb)*idgrd(xx,yy,zz+bb)
                        + idgrd(xx,yy+aa,zz)*idgrd(xx,yy+aa,zz+bb)*idgrd(xx,yy,zz+bb);
              for (cc = -1 ; cc <= 1 ; cc += 2) {
                n3210[0] += idgrd(xx+aa,yy,zz)*idgrd(xx+aa,yy+bb,zz)*idgrd(xx+aa,yy,zz+cc)
                          * idgrd(xx+aa,yy+bb,zz+cc)*idgrd(xx,yy+bb,zz)
                          * idgrd(xx,yy+bb,zz+cc)*idgrd(xx,yy,zz+cc);
              }
            }
          }
        }
      }
    }
  }
  freebin(&DGRID,nbins);
}

int eul3d(int* nn) {
  printf("3: %d 2: %d 1: %d 0: %d\n",nn[3],nn[2],nn[1],nn[0]);
  return -nn[3]+nn[2]-nn[1]+nn[0];
}
real vol3d(int* nn, real edgelength) {
  return nn[3]*pow(edgelength,3);
}
real sur3d(int* nn, real edgelength) {
  return (-6*nn[3]+2*nn[2])*pow(edgelength,2);
}
int twoB3d(int* nn, real edgelength) {
  return (3*nn[3]-2*nn[2]+nn[1])*edgelength;
}

bool neighbour_bool(ivec ind1, ivec ind2, ivec dim) {
  int i,nb;
  nb=0;
  for (i = 0 ; i < DIM ; i++) {
    if (ind1[i] == ind2[i])
      nb+=10;
    else if (abs(ind1[i]-ind2[i]) == 1 || abs(ind1[i]-ind2[i]) == dim[i]-1)
      nb+=1;
    else 
      break;
  }
  return (nb == 21);
}

bool ImgMan(ivec* coord, int**** image, ivec nbins, int** offsets) {
//ELIMINATE BUMPS/DENTS, CHANGE coord TO COORDINATES OF SINGLE (SIMILAR) NEIGHBOUR
  int val, nval, i, j, k, count;
  i = (*coord)[0]; j = (*coord)[1]; k = (*coord)[2];
  //val = (*image)[(*coord)[0]][(*coord)[1]][(*coord)[2]];
  //val = (*image)[i][j][k];
  val = pbc_array(*image,i,j,k,nbins,offsets);
  nval = labs(val-1);
  count=0;
  if(pbc_array(*image,i+1,j,k,nbins,offsets)==val) {
    count++;
    (*coord)[0]=i+1;
  }
  if(pbc_array(*image,i-1,j,k,nbins,offsets)==val) {
    count++;
    if (count==1) { (*coord)[0]=i-1; }
    else { (*coord)[0]=-23; return FALSE; } //leave this line in all cases... (except first)
  }
  if(pbc_array(*image,i,j+1,k,nbins,offsets)==val) {
    count++;
    if (count==1) { (*coord)[1]=j+1; }
    else { (*coord)[0]=-23; return FALSE; } //leave this line in all cases... (except first)
  }
  if(pbc_array(*image,i,j-1,k,nbins,offsets)==val) {
    count++;
    if (count==1) { (*coord)[1]=j-1; }
    else { (*coord)[0]=-23; return FALSE; } //leave this line in all cases... (except first)
  }
  if(pbc_array(*image,i,j,k+1,nbins,offsets)==val) {
    count++;
    if (count==1) { (*coord)[2]=k+1; }
    else { (*coord)[0]=-23; return FALSE; } //leave this line in all cases... (except first)
  }
  if(pbc_array(*image,i,j,k-1,nbins,offsets)==val) {
    count++;
    if (count==1) { (*coord)[2]=k-1; }
    else { (*coord)[0]=-23; return FALSE; } //leave this line in all cases... (except first)
  }
  if (count==0) {(*coord)[0]=-23; return FALSE; }
  else {
    pbc_array_w(image,i,j,k,nbins,offsets,nval); //invert pixel
    return TRUE;
  }
}

void cluster_filter(int**** image, ivec nbins, ivec* ndx, int size, int** offsets) {
  int i,j,k,a,b,c;
  int** cluster; //0 will be the offset of the last member + 1, that is the length of the array
  //MAYBE ALLOCATE CLUSTERS ONLY FULLY, IF THEY ARE A "EATER"??? MEMORY MF
  snew(cluster,ndx[0][0]);
  for (i = 0 ; i < ndx[0][0] ; i++) {
    snew(cluster[i],/*ndx[0][0]-i+2*/3); //do not yet allocate...
    cluster[i][0]=1;
    cluster[i][1]=i;
    cluster[i][2]=-22;
  }
#ifdef __OLD_PBC
#define img(s,t,u) (*image)[(s+nbins[0])%nbins[0]][(t+nbins[1])%nbins[1]][(u+nbins[2])%nbins[2]]
#else
#define img(s,t,u) pbc_array((*image),s,t,u,nbins,offsets)
#define wimg(s,t,u,v) pbc_array_w((image),(s),(t),(u),nbins,offsets,(v))
#endif


  //find clusters...
  int clusternumber;
  //clusternumber = ndx[0][0];
  clusternumber = ndx[0][0]-1; //MF NEWCHANGE added -1!!! ndx[0][0] = #x*#y*#z+1 [7.2.2008]
        fprintf(stderr,"initial clusternumber %d\n",clusternumber);
  for (i = 0 ; i < ndx[0][0]-1 ; i++) { //MF NEWCHANGE added -1
    (*image)[ndx[i+1][0]][ndx[i+1][1]][ndx[i+1][2]]=0;//temp. stored the backgrnd index there, need 0 for getn3210
                                                      //this one is a eater...
    for (a = 1 ; a <= cluster[i][0] ; a++) { //necessary if cluster eats another one 
      if (cluster[i][2]==-22) { //potential eater <== cluster[i][0] != 0
        fprintf(stderr,"got a -22! MF22\n");
        sfree(cluster[i]);              //free and
        snew(cluster[i],ndx[0][0]-i+2); //allocate again (this time for worst case)
        cluster[i][0]=1;                //you might want to use realloc instead...
        cluster[i][1]=i;
      }

        if (img(ndx[cluster[i][a]+1][0]+1,ndx[cluster[i][a]+1][1],ndx[cluster[i][a]+1][2])<0) {
                                                        //eaten ones/ex-eaters: 0! uneaten ones: negative!
          //eat it!! - increase length of i by one and copy its ndx into i
          cluster[i][0]++;
          cluster[i][cluster[i][0]]=cluster[-img(ndx[cluster[i][a]+1][0]+1,ndx[cluster[i][a]+1][1],ndx[cluster[i][a]+1][2])][1];
          //and eliminate it
          cluster[-img(ndx[cluster[i][a]+1][0]+1,ndx[cluster[i][a]+1][1],ndx[cluster[i][a]+1][2])][0]=0;
          //and make a note of eater in it
          clusternumber--;
#ifdef __OLD_PBC
          img(ndx[cluster[i][a]+1][0]+1,ndx[cluster[i][a]+1][1],ndx[cluster[i][a]+1][2])=0; //mark as eaten
#else
          wimg(ndx[cluster[i][a]+1][0]+1,ndx[cluster[i][a]+1][1],ndx[cluster[i][a]+1][2],0); //mark as eaten
#endif
        }
        if (img(ndx[cluster[i][a]+1][0]-1,ndx[cluster[i][a]+1][1],ndx[cluster[i][a]+1][2])<0) {
          //eat it!! - increase length of i by one and copy its ndx into i
          cluster[i][0]++;
          cluster[i][cluster[i][0]]=cluster[-img(ndx[cluster[i][a]+1][0]-1,ndx[cluster[i][a]+1][1],ndx[cluster[i][a]+1][2])][1];
          //and eliminate it
          cluster[-img(ndx[cluster[i][a]+1][0]-1,ndx[cluster[i][a]+1][1],ndx[cluster[i][a]+1][2])][0]=0;
          clusternumber--;
#ifdef __OLD_PBC
          img(ndx[cluster[i][a]+1][0]-1,ndx[cluster[i][a]+1][1],ndx[cluster[i][a]+1][2])=0; //mark as eaten
#else
          wimg(ndx[cluster[i][a]+1][0]-1,ndx[cluster[i][a]+1][1],ndx[cluster[i][a]+1][2],0); //mark as eaten
#endif
        }
        if (img(ndx[cluster[i][a]+1][0],ndx[cluster[i][a]+1][1]+1,ndx[cluster[i][a]+1][2])<0) {
          //eat it!! - increase length of i by one and copy its ndx into i
          cluster[i][0]++;
          cluster[i][cluster[i][0]]=cluster[-img(ndx[cluster[i][a]+1][0],ndx[cluster[i][a]+1][1]+1,ndx[cluster[i][a]+1][2])][1];
          //and eliminate it
          cluster[-img(ndx[cluster[i][a]+1][0],ndx[cluster[i][a]+1][1]+1,ndx[cluster[i][a]+1][2])][0]=0;
          clusternumber--;
#ifdef __OLD_PBC
          img(ndx[cluster[i][a]+1][0],ndx[cluster[i][a]+1][1]+1,ndx[cluster[i][a]+1][2])=0; //mark as eaten
#else
          wimg(ndx[cluster[i][a]+1][0],ndx[cluster[i][a]+1][1]+1,ndx[cluster[i][a]+1][2],0); //mark as eaten
#endif
        }
        if (img(ndx[cluster[i][a]+1][0],ndx[cluster[i][a]+1][1]-1,ndx[cluster[i][a]+1][2])<0) {
          //eat it!! - increase length of i by one and copy its ndx into i
          cluster[i][0]++;
          cluster[i][cluster[i][0]]=cluster[-img(ndx[cluster[i][a]+1][0],ndx[cluster[i][a]+1][1]-1,ndx[cluster[i][a]+1][2])][1];
          //and eliminate it
          cluster[-img(ndx[cluster[i][a]+1][0],ndx[cluster[i][a]+1][1]-1,ndx[cluster[i][a]+1][2])][0]=0;
          clusternumber--;
#ifdef __OLD_PBC
          img(ndx[cluster[i][a]+1][0],ndx[cluster[i][a]+1][1]-1,ndx[cluster[i][a]+1][2])=0; //mark as eaten
#else
          wimg(ndx[cluster[i][a]+1][0],ndx[cluster[i][a]+1][1]-1,ndx[cluster[i][a]+1][2],0); //mark as eaten
#endif
        }
        if (img(ndx[cluster[i][a]+1][0],ndx[cluster[i][a]+1][1],ndx[cluster[i][a]+1][2]+1)<0) {
          //eat it!! - increase length of i by one and copy its ndx into i
          cluster[i][0]++;
          cluster[i][cluster[i][0]]=cluster[-img(ndx[cluster[i][a]+1][0],ndx[cluster[i][a]+1][1],ndx[cluster[i][a]+1][2]+1)][1];
          //and eliminate it
          cluster[-img(ndx[cluster[i][a]+1][0],ndx[cluster[i][a]+1][1],ndx[cluster[i][a]+1][2]+1)][0]=0;
          clusternumber--;
#ifdef __OLD_PBC
          img(ndx[cluster[i][a]+1][0],ndx[cluster[i][a]+1][1],ndx[cluster[i][a]+1][2]+1)=0; //mark as eaten
#else
          wimg(ndx[cluster[i][a]+1][0],ndx[cluster[i][a]+1][1],ndx[cluster[i][a]+1][2]+1,0); //mark as eaten
#endif
        }
        if (img(ndx[cluster[i][a]+1][0],ndx[cluster[i][a]+1][1],ndx[cluster[i][a]+1][2]-1)<0) {
          //eat it!! - increase length of i by one and copy its ndx into i
          cluster[i][0]++;
          cluster[i][cluster[i][0]]=cluster[-img(ndx[cluster[i][a]+1][0],ndx[cluster[i][a]+1][1],ndx[cluster[i][a]+1][2]-1)][1];
          //and eliminate it
          cluster[-img(ndx[cluster[i][a]+1][0],ndx[cluster[i][a]+1][1],ndx[cluster[i][a]+1][2]-1)][0]=0;
          clusternumber--;
#ifdef __OLD_PBC
          img(ndx[cluster[i][a]+1][0],ndx[cluster[i][a]+1][1],ndx[cluster[i][a]+1][2]-1)=0; //mark as eaten
#else
          wimg(ndx[cluster[i][a]+1][0],ndx[cluster[i][a]+1][1],ndx[cluster[i][a]+1][2]-1,0); //mark as eaten
#endif
        }


          fprintf(stderr,"end of eating loop, clusternumber: %d\n",clusternumber); //MF DEBUG
    }
  }

          printf("clusternumber = %d ndx[0][0] = %d\n",clusternumber,ndx[0][0]);
  //...and filter clusters smaller than "size"
  for (i = 0 ; i < ndx[0][0]-1 ; i++) { //MF because ndx[0][0] starts at 1 and is increased by one per bgrd-pixel
                                        //the index of the last entry in the actual list (starts at 0) is ndx[0][0]-2
    if  (cluster[i][0]>0) printf("cluster[%d][0] = %d\n",i,cluster[i][0]);
    if (cluster[i][0]>0 && cluster[i][0]<=size) {
      for (j = 1 ; j <= cluster[i][0] ; j++) {
        (*image)[ndx[cluster[i][j]+1][0]][ndx[cluster[i][j]+1][1]][ndx[cluster[i][j]+1][2]]=1;
          printf("still filtering x:%d y:%d z:%d\n",ndx[cluster[i][j]+1][0],ndx[cluster[i][j]+1][1],ndx[cluster[i][j]+1][2]); //MF DEBUG
      }
    }
  }
  for (i = 0 ; i < ndx[0][0] ; i++) {
    sfree(cluster[i]);
  }
  sfree(cluster);
}

int gmx_mia(int argc,char *argv[])
{
  static char *desc[] = {
    "This is a tool for the morphological analysis of ",
    "binary mixtures. Use this executable to obtain the ",
    "global values of the Minkowski functionals. ",
    "For the local curvatures, use \"g_mialoc\" instead.",
  };
  
  static real bindim=0.5;
  static real radius=0.235;
  static int np=1;
  static int nlc=2;
  static bool LV=0;
  static int cs=1;
  static int npts=27;
  static int thresh=0;
  //static char  *res_name1="W",*res_name2="Na",*res_name3="Cl";
  //char ** negphase;
  //snew(negphase,3); negphase[0]=res_name1;  negphase[1]=res_name2; negphase[2]=res_name3;//free after phasedetermination!
  int i,j,k; //counters for many things, leave them!
  int *** bin;   //grid: used first for counting atom-dots in the cell and then converted to b/w PIC
  int *** locID;   //grid: stores identifier for local curvatures
  ivec* grdndx; //grdndx = NULL;
  bool VISIBILI = FALSE;

  /* Extra arguments - but note how you always get the begin/end
   * options when running the program, without mentioning them here!
   */
  
  t_pargs pa[] = {
    { "-np", FALSE, etINT,  {&np},         "Specifies wether to treat the negative phase explicitely (2) or not (1)"},
    { "-locval", FALSE, etBOOL,  {&LV},    "Compute local curvatures (1) or not (0)?"},
    { "-nlc", FALSE, etINT,  {&nlc},       "Averaging is performed over neighbouring surface cells up to nlc cells away"},
    { "-cs"  , FALSE, etINT,  {&cs},       "Maximum number of clustered background grids that are filtered"},
    { "-npts", FALSE, etINT,  {&npts},     "Number of points used for filling of sphere"},
    { "-thresh1", FALSE, etINT,  {&thresh}, "Threshold: minimum number of points required to count a grid cell as positive"},
    { "-sr"  , FALSE, etREAL, {&radius},   "Radius of spheres used for beads in nm"},
    { "-dim" , FALSE, etREAL, {&bindim},   "Edge length of cubes for grid in nm"}
  };
  
  //t_topology top;
  char       title[STRLEN];
  t_trxframe fr;
  //rvec       *xtop;
  matrix     box;
  int        status;
  int        flags = TRX_READ_X;
  //bool       *phaselist; //exactly that
  int        atnr;  //counter
  ivec       nbins; //dimension of bin/GRID/DGRID
  int**      offsets; //offsets of periodic image in triclinic boxes
  int*       n3210; //number of cubes, faces, edges, vertices
  snew(n3210,4);
  //int*     oct3210; //number of cubes, faces, edges, vertices used to identify octant for local Curvatures
  //snew(oct3210,4);
  t_atoms atoms;
  atom_id    **index;    //numbers of index file
  int        *nindex;    //length of groups in index
  char       **indexgrp; //name of index groups - not used, but required for formal reasons...
  ivec       tmpvec;
  real       *global_curv;
  snew(global_curv,5);

  t_filenm fnm[] = {
    //{ efTPS,  NULL,  NULL, ffREAD },   /* this is for the topology     */
    { efTRX, "-f", NULL, ffREAD },     /* and this for the trajectory  */
    { efNDX, "-n", NULL, ffREAD },     /* and this for the index  */
    { efXVG,"-o","mia",ffWRITE }       /* xvgr output file 	       */
  };
  
#define NFILE asize(fnm)

  CopyRight(stderr,argv[0]);

  /* This is the routine responsible for adding default options,
   * calling the X/motif interface, etc. */
  parse_common_args(&argc,argv,PCA_CAN_TIME | PCA_CAN_VIEW,
		    NFILE,fnm,asize(pa),pa,asize(desc),desc,0,NULL);

  /* We don't need any topology information to write the coordinates,
   * but to show how it works we start by writing the name and
   * charge of the selected atom. It returns a boolean telling us
   * whether the topology was found and could be read
   */
  
  //read_tps_conf(ftp2fn(efTPS,NFILE,fnm),title,&top,&xtop,NULL,box,TRUE);
  //sfree(xtop);

  //open xvg
  FILE* fp2;
  if (VISIBILI) {
    char* fn2 = "visual.py";
    fp2 = fopen(fn2,"w");
  }
  FILE* fp;
  fp = xvgropen(opt2fn("-o",NFILE,fnm),"MIA","Time [ps]","Minkowski Functionals");
  //Make Phaselist!
  //phaselist = makePhaseList(top,negphase,nres);

  
  /* The first time we read data is a little special */
  read_first_frame(&status,ftp2fn(efTRX,NFILE,fnm),&fr,flags);

  snew(nindex,np);
  snew(index,np);
  snew(indexgrp,np);
  get_index(&atoms,ftp2fn_null(efNDX,NFILE,fnm),np,nindex,index,indexgrp);
   
  /* This is the main loop over frames */
  do {
    /* coordinates are available in the vector fr.x
     * you can find this and all other structures in
     * the types directory under the gromacs include dir.
     * Note how flags determines wheter to read x/v/f!
     */
    //printf("Coordinates at t=%8.3f : %8.5f %8.5f %8.5f\n",fr.time,fr.x[1][XX],fr.x[1][YY],fr.x[1][ZZ]);
    for (i=0;i<5;i++)
      global_curv[i]=0;

    //NEW AND BETTER ALL IN ONE BINNING FUNCTION!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
    bin = binning(fr, index, nindex, np, bindim, &nbins, &npts, radius, &grdndx,nlc,LV,thresh,cs,&offsets);

          fprintf(stderr,"done binning\n");
    
    //apply Cluster Filter
    if (cs>0) {  
      cluster_filter(&bin, nbins, grdndx, cs, offsets);
          fprintf(stderr,"done filtering\n");
    }

    //IMAGE MANIPULATION
    for (i = 0 ; i<nbins[0] ; i++) {
      for (j = 0 ; j<nbins[1] ; j++) {
        for (k = 0 ; k<nbins[2] ; k++) {
          tmpvec[0]=i;tmpvec[1]=j;tmpvec[2]=k;
          while(tmpvec[0]!=-23) {
            ImgMan(&tmpvec, &bin, nbins, offsets);
          }
        }
      }
    }
    fprintf(stderr,"done IMANing\n");

    /* MF TEST
    for (i = 0 ; i<nbins[0] ; i++) {
      for (j = 0 ; j<nbins[1] ; j++) {
        for (k = 0 ; k<nbins[2] ; k++) {
          //printf("bin[%d][%d][%d]=%d\n",i,j,k,bin[i][j][k]);
          if (j==3) {
            bin[i][j][k]=1;
          } else {
            bin[i][j][k]=0;
          }
        }
      }
    }
    //bin[3][3][3]=0;
    //bin[7][7][8]=1;
    */

    //PRINT VISUALISATION IF DESIRED
    if (VISIBILI) {
      write_cgo_array(fp2,bin,nbins,bindim);
    }

    //COUNT CUBES, FACES, EDGES AND VERTICES!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
    getn3210(n3210,nbins,bin,offsets);
          fprintf(stderr,"done getn\n");


    //NOW FOR THE LOCAL CURVATURES!!!!!!!!!!!!!
    if (LV) {
      locID = get_locID(bin,nbins,nlc,bindim,offsets,global_curv);

      freebin(&locID,nbins);
    }

    for (i=0 ; i<3 ; i++) {
      sfree(offsets[i]);
    }
    sfree(offsets);
    freebin(&bin,nbins);
    sfree(grdndx);
    
    //print 2 file (printMIA)
    fprintf(fp,"%10g %10g %10g %10g %10d %g %g %g\n",fr.time,vol3d(n3210,bindim),sur3d(n3210,bindim),M_PI*twoB3d(n3210,bindim),eul3d(n3210),global_curv[2],global_curv[0],global_curv[1]/2/M_PI);
    printf("GLOBAL Mean:%f Gauss:%f Area: %f\n",global_curv[0],global_curv[1],global_curv[2]);
    printf("alt Mean:%f Gauss:%f\n",global_curv[3],global_curv[4]);

  } while(read_next_frame(status,&fr));

  sfree(global_curv);

  fclose(fp);
  if (VISIBILI) {
    fclose(fp2);
  }
  sfree(n3210);
  //sfree(oct3210);
  thanx(stderr);
  
  return 0;
}

