/*
 * $Id: template.c,v 1.4 2001/07/23 15:28:29 lindahl Exp $
 * 
 *                This source code is part of
 * 
 *                 G   R   O   M   A   C   S
 * 
 *          GROningen MAchine for Chemical Simulations
 * 
 *                        VERSION 3.0
 * 
 * Copyright (c) 1991-2001
 * BIOSON Research Institute, Dept. of Biophysical Chemistry
 * University of Groningen, The Netherlands
 * 
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 * 
 * If you want to redistribute modifications, please consider that
 * scientific software is very special. Version control is crucial -
 * bugs must be traceable. We will be happy to consider code for
 * inclusion in the official distribution, but derived work must not
 * be called official GROMACS. Details are found in the README & COPYING
 * files - if they are missing, get the official version at www.gromacs.org.
 * 
 * To help us fund GROMACS development, we humbly ask that you cite
 * the papers on the package - you can find them in the top README file.
 * 
 * Do check out http://www.gromacs.org , or mail us at gromacs@gromacs.org .
 * 
 * And Hey:
 * Gyas ROwers Mature At Cryogenic Speed
 */

/* This line is only for CVS version info */
static char *SRCID_template_c = "$Id: template.c,v 1.4 2001/07/23 15:28:29 lindahl Exp $";

#include "statutil.h"
#include "typedefs.h"
#include "smalloc.h"
#include "vec.h"
#include "copyrite.h"
#include "statutil.h"
#include "tpxio.h"
#include "string2.h"
#include "pbc.h"
#include "xvgr.h"
#include "math.h"
#include "physics.h" //for M_PI ...
#include "index.h"

/*#define __OLD_PBC */
/*#define _TEST_ */

#define grd(i,j,k) GRID[i][j][k]
#define dgrd(i,j,k) DGRID[i][j][k]
#ifdef __OLD_PBC
#define idgrd(i,j,k) (1-DGRID[(i+nbins[0])%nbins[0]][(j+nbins[1])%nbins[1]][(k+nbins[2])%nbins[2]])
#else
#define idgrd(i,j,k) (1-pbc_array(DGRID,i,j,k,nbins,offsets))
#endif

void getn3210(int*,ivec,int***,int**);

void freebin(int****, ivec);

void rpbc_array_w(real**** array, int xxx, int yyy, int zzz, ivec nbins, int ** offsets, real value) {
//writes into array respecting PBC and possible offsets present in a triclinic system
  /*
  expected format of offsets:
  0  x-offset of b  x-offset of c
  0  0              y-offset of c 
  0  0              0
  */
  ivec coord;
  coord[0]=xxx;coord[1]=yyy;coord[2]=zzz;
  int i,j;
  for (i = 2 ; i >= 0 ; i--) {
    if (coord[i] >= nbins[i]) {
      coord[i]-=nbins[i];
      for (j = i-1 ; j >= 0 ; j--) {
        coord[j]-=offsets[i][j]; //+?

      }
    }
    else if (coord[i] < 0) {
      coord[i]+=nbins[i];
      for (j = i-1 ; j >= 0 ; j--) {
        coord[j]+=offsets[i][j]; //-?

      }
    }
  }
  (*array)[coord[0]][coord[1]][coord[2]]=value;
}

void dpbc_array_w(double**** array, int xxx, int yyy, int zzz, ivec nbins, int ** offsets, double value) {
//writes into array respecting PBC and possible offsets present in a triclinic system
  /*
  expected format of offsets:
  0  x-offset of b  x-offset of c
  0  0              y-offset of c 
  0  0              0
  */
  ivec coord;
  coord[0]=xxx;coord[1]=yyy;coord[2]=zzz;
  int i,j;
  for (i = 2 ; i >= 0 ; i--) {
    if (coord[i] >= nbins[i]) {
      coord[i]-=nbins[i];
      for (j = i-1 ; j >= 0 ; j--) {
        coord[j]-=offsets[i][j]; //+?
      }
    }
    else if (coord[i] < 0) {
      coord[i]+=nbins[i];
      for (j = i-1 ; j >= 0 ; j--) {
        coord[j]+=offsets[i][j]; //-?
      }
    }
  }
  (*array)[coord[0]][coord[1]][coord[2]]=value;
}

void pbc_array_w(int**** array, int xxx, int yyy, int zzz, ivec nbins, int ** offsets, int value) {
//writes into array respecting PBC and possible offsets present in a triclinic system
  /*
  expected format of offsets:
  0  x-offset of b  x-offset of c
  0  0              y-offset of c 
  0  0              0
  */
  ivec coord;
  coord[0]=xxx;coord[1]=yyy;coord[2]=zzz;
  int i,j;
  for (i = 2 ; i >= 0 ; i--) {
    if (coord[i] >= nbins[i]) {
      coord[i]-=nbins[i];
      for (j = i-1 ; j >= 0 ; j--) {
        coord[j]-=offsets[i][j]; //+?
      }
    }
    else if (coord[i] < 0) {
      coord[i]+=nbins[i];
      for (j = i-1 ; j >= 0 ; j--) {
        coord[j]+=offsets[i][j]; //-?
      }
    }
  }
  (*array)[coord[0]][coord[1]][coord[2]]=value;
}

real rpbc_array(real*** array, int xxx, int yyy, int zzz, ivec nbins, int ** offsets) {
//adresses array respecting PBC and possible offsets present in a triclinic system
  /*
  expected format of offsets:
  0  x-offset of b  x-offset of c
  0  0              y-offset of c 
  0  0              0
  */
  ivec coord;
  coord[0]=xxx;coord[1]=yyy;coord[2]=zzz;
  int i,j;
  for (i = 2 ; i >= 0 ; i--) {
    if (coord[i] >= nbins[i]) {
      coord[i]-=nbins[i];
      for (j = i-1 ; j >= 0 ; j--) {
        coord[j]-=offsets[i][j]; //+?
      }
    }
    else if (coord[i] < 0) {
      coord[i]+=nbins[i];
      for (j = i-1 ; j >= 0 ; j--) {
        coord[j]+=offsets[i][j]; //-?
      }
    }
  }
  return array[coord[0]][coord[1]][coord[2]];
}

double dpbc_array(double*** array, int xxx, int yyy, int zzz, ivec nbins, int ** offsets) {
//adresses array respecting PBC and possible offsets present in a triclinic system
  /*
  expected format of offsets:
  0  x-offset of b  x-offset of c
  0  0              y-offset of c 
  0  0              0
  */
  ivec coord;
  coord[0]=xxx;coord[1]=yyy;coord[2]=zzz;
  int i,j;
  for (i = 2 ; i >= 0 ; i--) {
    if (coord[i] >= nbins[i]) {
      coord[i]-=nbins[i];
      for (j = i-1 ; j >= 0 ; j--) {
        coord[j]-=offsets[i][j]; //+?
      }
    }
    else if (coord[i] < 0) {
      coord[i]+=nbins[i];
      for (j = i-1 ; j >= 0 ; j--) {
        coord[j]+=offsets[i][j]; //-?
      }
    }
  }
  return array[coord[0]][coord[1]][coord[2]];
}

int pbc_array(int*** array, int xxx, int yyy, int zzz, ivec nbins, int ** offsets) {
//adresses array respecting PBC and possible offsets present in a triclinic system
  /*
  expected format of offsets:
  0  x-offset of b  x-offset of c
  0  0              y-offset of c 
  0  0              0
  */
  ivec coord;
  coord[0]=xxx;coord[1]=yyy;coord[2]=zzz;
  int i,j;
  for (i = 2 ; i >= 0 ; i--) {
    if (coord[i] >= nbins[i]) {
      coord[i]-=nbins[i];
      for (j = i-1 ; j >= 0 ; j--) {
        coord[j]-=offsets[i][j]; //+?
      }
    }
    else if (coord[i] < 0) {
      coord[i]+=nbins[i];
      for (j = i-1 ; j >= 0 ; j--) {
        coord[j]+=offsets[i][j]; //-?
      }
    }
  }
  return array[coord[0]][coord[1]][coord[2]];
}

void gen_rot_matrices(matrix Xr, matrix Yr, matrix Zr, real phi, real epsilon, real theta) {
  int i,j;
  for (i=0;i<3;i++) {
    for (j=0;j<3;j++) {
      Xr[i][j]=0;
      Yr[i][j]=0;
      Zr[i][j]=0;
    }
  }
  Xr[0][0]=1; Xr[1][1]=cos(phi); Xr[2][2]=cos(phi); Xr[1][2]=sin(phi); Xr[2][1]=-sin(phi);
  Yr[1][1]=1; Yr[0][0]=cos(epsilon); Yr[2][2]=cos(epsilon); Yr[2][0]=sin(epsilon); Yr[0][2]=-sin(epsilon);
  Zr[2][2]=1; Zr[1][1]=cos(theta); Zr[0][0]=cos(theta); Zr[0][1]=sin(theta); Zr[1][0]=-sin(theta);
}

void put_atoms_in_box_and_make_offset_multpl_dim(matrix box,int natoms,rvec x[],real dim)
{
  int i,m,d,j;
  matrix shift;
  for (i=1 ; i<=2 ; i++) {
    for (j=0 ; j<i ; j++) {
      shift[i][j] = box[i][j]-dim*(int)(box[i][j]/dim);
    }
  }

  for(i=0; (i<natoms); i++) {
    for(m=DIM-1; m>=0; m--) {
      while (x[i][m] < 0) 
	for(d=0; d<=m; d++)       //Starting with Z (m=2), the 2 whiles shift the atom by one boxvector box[m],
	  x[i][d] += box[m][d];   //changing the coordinates up to and including m
      while (x[i][m] >= box[m][m])
	for(d=0; d<=m; d++)
	  x[i][d] -= box[m][d];
      for (j=0 ; j<m ; j++) {                   
        x[i][j]-=shift[m][j]*x[i][m]/box[m][m]; //shifts lower coordinates for the current m (e.g. X and Y for Z)
      }                                         //doing this after the 2 whiles ensures, x[i][m] is in [0/box[m][m][
    }
  }
}

void write_cgo_frame( FILE *fp, real dx, real dy, real dz, real r )
{
    char *CGO_CYLINDER = "CYLINDER, %8.3f, %8.3f, %8.3f, %8.3f, %8.3f, %8.3f, %8.3f, 1, 1, 1, 1, 1, 1, \n";

    fprintf( fp, CGO_CYLINDER, 0.0, 0.0, 0.0,  dx, 0.0, 0.0, r );
    fprintf( fp, CGO_CYLINDER, 0.0, 0.0, 0.0, 0.0,  dy, 0.0, r );
    fprintf( fp, CGO_CYLINDER, 0.0, 0.0, 0.0, 0.0, 0.0,  dz, r );
    fprintf( fp, CGO_CYLINDER,  dx, 0.0, 0.0,  dx,  dy, 0.0, r );
    fprintf( fp, CGO_CYLINDER,  dx, 0.0, 0.0,  dx, 0.0,  dz, r );
    fprintf( fp, CGO_CYLINDER, 0.0,  dy, 0.0,  dx,  dy, 0.0, r );
    fprintf( fp, CGO_CYLINDER, 0.0,  dy, 0.0, 0.0,  dy,  dz, r );
    fprintf( fp, CGO_CYLINDER, 0.0, 0.0,  dz,  dx, 0.0,  dz, r );
    fprintf( fp, CGO_CYLINDER, 0.0, 0.0,  dz, 0.0,  dy,  dz, r );
    fprintf( fp, CGO_CYLINDER,  dx,  dy, 0.0,  dx,  dy,  dz, r );
    fprintf( fp, CGO_CYLINDER,  dx, 0.0,  dz,  dx,  dy,  dz, r );
    fprintf( fp, CGO_CYLINDER, 0.0,  dy,  dz,  dx,  dy,  dz, r );
}

void write_cgo_triangle( FILE *fp, real nx, real ny, real nz, real ax, real ay, real az, real bx, real by, real bz, real cx, real cy, real cz )
{
    fprintf( fp, "NORMAL, %8.3f, %8.3f, %8.3f, ", nx, ny, nz );
    fprintf( fp, "VERTEX, %8.3f, %8.3f, %8.3f, ", ax, ay, az );
    fprintf( fp, "VERTEX, %8.3f, %8.3f, %8.3f, ", bx, by, bz );
    fprintf( fp, "VERTEX, %8.3f, %8.3f, %8.3f, \n", cx, cy, cz );
}

void write_cgo_cube( FILE *fp, real x0, real y0, real z0, real d )
{
    real x1,y1,z1;

    x1 = x0 + d;
    y1 = y0 + d;
    z1 = z0 + d;

    fprintf( fp, "BEGIN, TRIANGLES, COLOR, 1, 1, 1, \n" );

    write_cgo_triangle( fp, 1, 0, 0, x1, y0, z0, x1, y0, z1, x1, y1, z0 );
    write_cgo_triangle( fp, 1, 0, 0, x1, y0, z1, x1, y1, z0, x1, y1, z1 );
    write_cgo_triangle( fp,-1, 0, 0, x0, y0, z0, x0, y0, z1, x0, y1, z0 );
    write_cgo_triangle( fp,-1, 0, 0, x0, y0, z1, x0, y1, z0, x0, y1, z1 );

    write_cgo_triangle( fp, 0, 1, 0, x0, y1, z0, x0, y1, z1, x1, y1, z0 );
    write_cgo_triangle( fp, 0, 1, 0, x0, y1, z1, x1, y1, z0, x1, y1, z1 );
    write_cgo_triangle( fp, 0,-1, 0, x0, y0, z0, x0, y0, z1, x1, y0, z0 );
    write_cgo_triangle( fp, 0,-1, 0, x0, y0, z1, x1, y0, z0, x1, y0, z1 );

    write_cgo_triangle( fp, 0, 0, 1, x0, y0, z1, x1, y0, z1, x0, y1, z1 );
    write_cgo_triangle( fp, 0, 0, 1, x0, y1, z1, x1, y0, z1, x1, y1, z1 );
    write_cgo_triangle( fp, 0, 0,-1, x0, y0, z0, x1, y0, z0, x0, y1, z0 );
    write_cgo_triangle( fp, 0, 0,-1, x0, y1, z0, x1, y0, z0, x1, y1, z0 );

    fprintf( fp, "END, \n" );
}
void write_color_cube( FILE *fp, real x0, real y0, real z0, real d, real color, /*for white 0 curv*/real curv)
{
    real x1,y1,z1;

    x1 = x0 + d;
    y1 = y0 + d;
    z1 = z0 + d;

    if (curv < 0) { //color [0.0/0.5[
      fprintf( fp, "BEGIN, TRIANGLES, COLOR, %3.2f, %3.2f, 1.00, \n",0.0+2*color,0.0+2*color );
      //if (color < 0 || color >= 0.5) { //MF DEBUG
      //  printf ("color range error: curv %14.12f color %3.2f \n",curv,color);
      //}
    } else { //color [0.5/1.0]
      fprintf( fp, "BEGIN, TRIANGLES, COLOR, 1.00, %3.2f, %3.2f, \n",1.0-2*(color-0.5),1.0-2*(color-0.5) );
      //if (color < 0.5 || color > 1.0) { //MF DEBUG
      //  printf ("color range error: curv %14.12f color %3.2f \n",curv,color);
      //}
    }

    write_cgo_triangle( fp, 1, 0, 0, x1, y0, z0, x1, y0, z1, x1, y1, z0 );
    write_cgo_triangle( fp, 1, 0, 0, x1, y0, z1, x1, y1, z0, x1, y1, z1 );
    write_cgo_triangle( fp,-1, 0, 0, x0, y0, z0, x0, y0, z1, x0, y1, z0 );
    write_cgo_triangle( fp,-1, 0, 0, x0, y0, z1, x0, y1, z0, x0, y1, z1 );

    write_cgo_triangle( fp, 0, 1, 0, x0, y1, z0, x0, y1, z1, x1, y1, z0 );
    write_cgo_triangle( fp, 0, 1, 0, x0, y1, z1, x1, y1, z0, x1, y1, z1 );
    write_cgo_triangle( fp, 0,-1, 0, x0, y0, z0, x0, y0, z1, x1, y0, z0 );
    write_cgo_triangle( fp, 0,-1, 0, x0, y0, z1, x1, y0, z0, x1, y0, z1 );

    write_cgo_triangle( fp, 0, 0, 1, x0, y0, z1, x1, y0, z1, x0, y1, z1 );
    write_cgo_triangle( fp, 0, 0, 1, x0, y1, z1, x1, y0, z1, x1, y1, z1 );
    write_cgo_triangle( fp, 0, 0,-1, x0, y0, z0, x1, y0, z0, x0, y1, z0 );
    write_cgo_triangle( fp, 0, 0,-1, x0, y1, z0, x1, y0, z0, x1, y1, z0 );

    fprintf( fp, "END, \n" );
}

void write_cgo_array( FILE *fp, int ***array, ivec bins, real d )
{
    int i,j,k;
    real dx,dy,dz;

    fprintf( fp, "from pymol.cgo import * \nfrom pymol import cmd \n\nobj = [ \n" );
    write_cgo_frame( fp, bins[0]*d, bins[1]*d, bins[2]*d, 0.1 );

    for ( i=0; i<bins[0]; i++ )
    {
	dx = i*d;
	for ( j=0; j<bins[1]; j++ )
	{
	    dy = j*d;
	    for ( k=0; k<bins[2]; k++ )
		if ( array[i][j][k] )
		{
		    dz = k*d;
		    write_cgo_cube( fp, dx, dy, dz, d );
		}
	}
    }
    
    fprintf( fp, "]\n\ncmd.load_cgo(obj, \"array3d\")\n" );
}
void write_color_image( FILE *fp, int ***image, ivec bins, real d, double *** curv, double scale, double min)
{
  int i,j,k;
  real dx,dy,dz;
  real color;

  fprintf( fp, "from pymol.cgo import * \nfrom pymol import cmd \n\nobj = [ \n" );
  write_cgo_frame( fp, bins[0]*d, bins[1]*d, bins[2]*d, 0.1 );

  for ( i=0; i<bins[0]; i++ )
  {
      dx = i*d;
      for ( j=0; j<bins[1]; j++ )
      {
          dy = j*d;
          for ( k=0; k<bins[2]; k++ )
      	if ( image[i][j][k] )
      	{
      	    dz = k*d;
            //compute color
            color=(curv[i][j][k]-min)*scale;
            if (color > 1.0) {
              color = 1.0;
              printf("WARNING (i) - color > 1.0: %18.16f\n",color);
            }
            else if (color < 0.0) {
              color = 0.0;
              printf("WARNING (i) - color < 0.0: %18.16f\n",color);
            }
      	    write_color_cube( fp, dx, dy, dz, d, color, curv[i][j][k]/*for white 0 curv*/);
      	}
      }
  }
  
  fprintf( fp, "]\n\ncmd.load_cgo(obj, \"array3d\")\n" );
}

void loc_color_image( FILE *fp, int ***count, ivec bins, real d, real *** curv, real scale, real min, real thresh)
{
    int i,j,k;
    real dx,dy,dz;
    real color;

    fprintf( fp, "from pymol.cgo import * \nfrom pymol import cmd \n\nobj = [ \n" );
    write_cgo_frame( fp, bins[0]*d, bins[1]*d, bins[2]*d, 0.1 );

    for ( i=0; i<bins[0]; i++ )
    {
	dx = i*d;
	for ( j=0; j<bins[1]; j++ )
	{
	    dy = j*d;
	    for ( k=0; k<bins[2]; k++ ) {
		if ( count[i][j][k] >= thresh )
		{
		    dz = k*d;
                    //compute color
                    color=(curv[i][j][k]-min)*scale;
                    if (color > 1.0) {
                      color = 1.0;
                      printf("WARNING - color > 1.0: %18.16f\n",color);
                    }
                    else if (color < 0.0) {
                      color = 0.0;
                      printf("WARNING - color < 0.0: %18.16f\n",color);
                    }
		    write_color_cube( fp, dx, dy, dz, d, color, curv[i][j][k]/*for white 0 curv*/);
		}
            }
	}
    }
    
    fprintf( fp, "]\n\ncmd.load_cgo(obj, \"array3d\")\n" );
}

void write_color_array( FILE *fp, double ***area, ivec bins, real d, double *** curv, double scale, double min)
{
    int i,j,k;
    real dx,dy,dz;
    real color;

    fprintf( fp, "from pymol.cgo import * \nfrom pymol import cmd \n\nobj = [ \n" );
    write_cgo_frame( fp, bins[0]*d, bins[1]*d, bins[2]*d, 0.1 );

    for ( i=0; i<bins[0]; i++ )
    {
	dx = i*d;
	for ( j=0; j<bins[1]; j++ )
	{
	    dy = j*d;
	    for ( k=0; k<bins[2]; k++ ) {
		if ( area[i][j][k] > 0.0 )
		{
		    dz = k*d;
                    //compute color
                    color=(curv[i][j][k]-min)*scale;
                    if (color > 1.0) {
                      color = 1.0;
                      printf("WARNING - color > 1.0: %18.16f\n",color);
                    }
                    else if (color < 0.0) {
                      color = 0.0;
                      printf("WARNING - color < 0.0: %18.16f\n",color);
                    }
		    write_color_cube( fp, dx, dy, dz, d, color, curv[i][j][k]/*for white 0 curv*/);
		}
            }
	}
    }
    
    fprintf( fp, "]\n\ncmd.load_cgo(obj, \"array3d\")\n" );
}


/* radius sphere, number of points, POINT in grid(real space), box */
rvec *probe_spheres(real sigma, int *npoints,rvec POINT,matrix box, real dim)
{ 
  //  printf("*npoints3 %d\n",*npoints);
  int LOOP=0,i,j,k,next=0;
  static rvec *Xp = NULL;
  static real dSIGMA =0, dTHETA=0, dPHI = 0;
  real r=0,theta=0, phi=0;

  rvec *Xtemp;
 
  real x,y,z;
  int ax1,ax2,ax3;


  x = POINT[XX];
  y = POINT[YY];
  z = POINT[ZZ];

  ax1=XX;
  ax2=YY;
  ax3=ZZ;

  snew(Xtemp,*npoints);


  /* this is only done the first time*/

  if(Xp == NULL)  //changed Xp to *Xp should be fine, but MF CHECK anyway! MEMORY
  { 
    printf("Xp == NULL\n"); //MF

 
    /*cubic root*/
    LOOP =(int)(pow(*npoints,(1.0/3.0))+0.1);

    /* redefine*/

    *npoints = LOOP*LOOP*LOOP;
    printf("*npoints4 %d\n",*npoints);

    snew(Xp,LOOP*LOOP*LOOP);

    dSIGMA = (real)sigma / LOOP;
    dTHETA = (real)(M_PI)/ LOOP;
    dPHI   = (real)(M_PI*2.0) / LOOP;

    next =0;
    for(i=0;i<LOOP;i++) //r-loop
    {
      //theta = 0; //MF
      phi = 0; //MF
      for(j=0;j<LOOP;j++) //phi-loop
      {
        theta = 0; //MF
        for(k=0;k<LOOP;k++) //theta-loop
        {

          Xp[next][ax1]= r*cos(theta)*sin(phi);
          Xp[next][ax2]= r*sin(theta)*sin(phi);
          Xp[next][ax3]= r*cos(phi);

          theta += dTHETA;
          next++;
        }
 
        phi += dPHI;

      }
      r += dSIGMA;

    }
  }

  /* now translate the reference sphere to the position of the atom*/
  /*put everyting in the box periodically !!!*/

  /*THIS IS THE ACTUAL MAKING OF THE SPHERE POINTS FOR THE INPUT COORDINATE*/
  for(i=0;i<*npoints;i++)
  {

    Xtemp[i][ax1] = Xp[i][ax1] + x;

  
    Xtemp[i][ax2] = Xp[i][ax2] + y;

 
    Xtemp[i][ax3] = Xp[i][ax3] + z;


  }
  //put_atoms_in_box(box,*npoints,Xtemp); //MF CHECK
  put_atoms_in_box_and_make_offset_multpl_dim(box,*npoints,Xtemp,dim); //MF CHECK


  /* now do the grid indexing over this array of *npoints*/
  return Xtemp;
}

rvec *probe_spheres_while_rotating(real sigma, int *npoints,rvec POINT,matrix box, real dim,matrix X_rot,matrix Y_rot,matrix Z_rot,rvec cntr)
{ 
  int LOOP=0,i,j,k,next=0;
  static rvec *Xp = NULL;
  static real dSIGMA =0, dTHETA=0, dPHI = 0;
  real r=0,theta=0, phi=0;

  rvec *Xtemp;
  rvec rotated,tmpPNT,tmp2PNT;
 
  real x,y,z;
  int ax1,ax2,ax3;

  //ROTATE HERE - THE ACTUAL ROTATION
  rvec_sub(POINT,cntr,tmpPNT);
  mvmul(X_rot,tmpPNT,tmp2PNT);
  mvmul(Y_rot,tmp2PNT,tmpPNT);
  mvmul(Z_rot,tmpPNT,rotated);
  rvec_inc(rotated,cntr);

  x = rotated[XX]; //MF correction
  y = rotated[YY];
  z = rotated[ZZ];

  ax1=XX;
  ax2=YY;
  ax3=ZZ;

  snew(Xtemp,*npoints);


  /* this is only done the first time*/

  if(Xp == NULL)  //changed Xp to *Xp should be fine, but MF CHECK anyway! MEMORY
  { 
    fprintf(stderr,"Xp == NULL\n"); //MF

 
    /*cubic root*/
    LOOP =(int)(pow(*npoints,(1.0/3.0))+0.1);

    /* redefine*/

    *npoints = LOOP*LOOP*LOOP;
    fprintf(stderr,"*npoints4 %d\n",*npoints);

    snew(Xp,LOOP*LOOP*LOOP);

    dSIGMA = (real)sigma / LOOP;
    dTHETA = (real)(M_PI)/ LOOP;
    dPHI   = (real)(M_PI*2.0) / LOOP;

    next =0;
    for(i=0;i<LOOP;i++) //r-loop
    {
      phi = 0; //MF
      for(j=0;j<LOOP;j++) //phi-loop
      {
        theta = 0; //MF
        for(k=0;k<LOOP;k++) //theta-loop
        {

          Xp[next][ax1]= r*cos(theta)*sin(phi);
          Xp[next][ax2]= r*sin(theta)*sin(phi);
          Xp[next][ax3]= r*cos(phi);

          theta += dTHETA;
          next++;
        }
 
        phi += dPHI;

      }
      r += dSIGMA;

    }
  }

  /* now translate the reference sphere to the position of the atom*/
  /*put everyting in the box periodically !!!*/

  /*THIS IS THE ACTUAL MAKING OF THE SPHERE POINTS FOR THE INPUT COORDINATE*/
  for(i=0;i<*npoints;i++)
  {

    Xtemp[i][ax1] = Xp[i][ax1] + x;
    if(Xtemp[i][ax1] >= box[ax1][ax1]) { //TO AVOID ROTATING THINGS OUT OF THE BOX
      Xtemp[0][0] = -9999.99;                 // SINCE PBC HAVE BEEN DESTROYED BY THE ROTATION
      break;
    }
    if(Xtemp[i][ax1] < 0) {
      Xtemp[0][0] = -9999.99;
      break;
    }
  
    Xtemp[i][ax2] = Xp[i][ax2] + y;
    if(Xtemp[i][ax2] >= box[ax2][ax2]) {
      Xtemp[0][0] = -9999.99;
      break;
    }
    if(Xtemp[i][ax2] < 0) {
      Xtemp[0][0] = -9999.99;
      break;
    }
 
    Xtemp[i][ax3] = Xp[i][ax3] + z;
    if(Xtemp[i][ax3] >= box[ax3][ax3]) {
      Xtemp[0][0] = -9999.99;
      break;
    }
    if(Xtemp[i][ax3] < 0) {
      Xtemp[0][0] = -9999.99;
      break;
    }

  }


  /* now do the grid indexing over this array of *npoints*/
  return Xtemp;
}

int*** get_locID(int*** bin ,ivec nbins,int nlc,real chi, int** offsets, real** loc_curv_bin, int *loc_curv_bin_length,matrix X_back, matrix Y_back, matrix Z_back,real distance, real* global_curv) {
  int nn,phase,i,j,k,plusminus,s,t,u;
  double chi2;
  double maxMean, minMean, maxGauss, minGauss;
  maxGauss=0.0; minGauss=0.0;
  chi2 = chi*chi;
  int*     oct3210; //number of cubes, faces, edges, vertices used to identify octant for local Curvatures
  snew(oct3210,4);
  plusminus = 1;
  ivec tmp;
  ivec noct;
  //ivec nbinsnlc;
  static real* cntr=NULL;
  if (cntr==NULL) {
    fprintf(stderr,"cntr == NULL\n"); //MF DEBUG
    snew(cntr,3);
    for(i=0;i<3;i++) {
      cntr[i]=nbins[i]*chi*0.5;
      printf("cntr[%d]=%f\n",i,cntr[i]);
    }
  }
  int** zeromatrix;
  snew(zeromatrix,3);
  for (i=0 ; i<3 ; i++) {
    printf("nbins[%d]=%d\n",i,nbins[i]);
    for (j=0;j<3;j++) {
      printf("offsets[%d][%d]=%d\n",i,j,offsets[i][j]);
    }
    noct[i]=3;
    snew(zeromatrix[i],3);
    for (j=0 ; j<3 ; j++) {
      zeromatrix[i][j]=0;
    }
  }
  //allocate bin
  int*** mybin;
  snew(mybin,(nbins)[0]);
  for (i = 0 ; i<(nbins)[0] ; i++) {
    snew(mybin[i],(nbins)[1]);
    for (j = 0 ; j<(nbins)[1] ; j++) {
      snew(mybin[i][j],(nbins)[2]); 
      /*TEST*/
      ///*
#ifdef  _TEST_
      for (k=0;k<nbins[2];k++) {
        if (FALSE /*j==5 || j==6 || j==4 || j==7*/) {
          bin[i][j][k]=1;
        } else {
          bin[i][j][k]=0;
        }
      }
      //*/
#endif
      /*TEST*/
    }
  }
  //TEST
#ifdef _TEST_
  ///*
  //bin[0][0][0]=1;
  //bin[0][1][2]=1;
  //bin[4][nbins[1]-1][4]=1;
  //bin[4][nbins[1]-2][4]=1;
  //bin[4][nbins[1]-3][4]=1;
  //bin[4+offsets[1][0]][0+0][4]=1;
  //bin[4+offsets[1][0]][0+1][4]=1;
  //bin[4+offsets[1][0]][0+2][4]=1;
  for(i=0;i<10;i++) {
    for(j=0;j<10;j++) {
      for(k=0;k<10;k++) {
      bin[4+i][4+j][nbins[2]-k-1]=1;
      //bin[4+i][4+j][nbins[2]-2]=1;
      //bin[4+i][4+j][nbins[2]-3]=1;
      bin[4+i+offsets[2][0]][4+j+offsets[2][1]][0+k]=1;
      //bin[4+i+offsets[2][0]][4+j+offsets[2][1]][0+1]=1;
      //bin[4+i+offsets[2][0]][4+j+offsets[2][1]][0+2]=1;
      }
    }
  }
  //*/
  /*
  bin[4][3][4]=1;
  bin[5][3][4]=1;
  bin[4][3][5]=1;
  bin[5][3][5]=1;
  bin[1][4][1]=0;
  bin[2][4][1]=0;
  bin[1][4][2]=0;
  bin[2][4][2]=0;
  bin[0][9][6]=1;
  bin[20][5][20]=0;
  bin[20][6][20]=0;
  bin[20][4][20]=0;
  bin[20][7][20]=0;
  bin[25][5][20]=0;
  bin[25][6][20]=0;
  bin[25][4][20]=0;
  bin[25][7][20]=0;
  bin[25][5][21]=0;
  bin[25][6][21]=0;
  bin[25][4][21]=0;
  bin[25][7][21]=0;
  bin[26][5][20]=0;
  bin[26][6][20]=0;
  bin[26][4][20]=0;
  bin[26][7][20]=0;
  bin[26][5][21]=0;
  bin[26][6][21]=0;
  bin[26][4][21]=0;
  bin[26][7][21]=0;

  bin[15][21][20]=1;
  bin[15][22][20]=1;
  bin[15][23][20]=1;
  bin[15][24][20]=1;
  bin[15][25][20]=1;
  bin[15][20][20]=1;

  bin[15][15][21]=1;
  bin[15][15][22]=1;
  bin[15][15][23]=1;
  bin[15][15][24]=1;
  bin[15][15][25]=1;
  bin[15][15][20]=1;

  bin[20][20][20]=1;
  bin[21][20][20]=1;
  bin[22][20][20]=1;
  bin[23][20][20]=1;
  bin[24][20][20]=1;
  bin[25][20][20]=1;

  bin[1][1][1]=1;
  */
#endif
  //TEST
  double*** meanbin,***gaussbin,***areabin;
  snew(gaussbin,nbins/*nlc*/[0]);     // it has already been assured that nbins is a multiple of nlc...
  snew(meanbin,nbins/*nlc*/[0]);
  snew(areabin,nbins/*nlc*/[0]);
  for (i = 0 ; i<nbins/*nlc*/[0] ; i++) {
    snew(gaussbin[i],nbins/*nlc*/[1]);
    snew(meanbin[i],nbins/*nlc*/[1]);
    snew(areabin[i],nbins/*nlc*/[1]);
    for (j = 0 ; j<nbins/*nlc*/[1] ; j++) {
      snew(gaussbin[i][j],nbins/*nlc*/[2]); 
      snew(meanbin[i][j],nbins/*nlc*/[2]); 
      snew(areabin[i][j],nbins/*nlc*/[2]); 
    }
  }
  int *** tmpoctant;
  snew(tmpoctant,noct[0]);
  for (i = 0 ; i<noct[0] ; i++) {
    snew(tmpoctant[i],noct[1]);
    for (j = 0 ; j<noct[1] ; j++) {
      snew(tmpoctant[i][j],noct[2]); 
    }
  }
  for (i = 0 ; i<noct[0] ; i++) {
    for (j = 0 ; j<noct[1] ; j++) {
      for (k = 0 ; k<noct[2] ; k++) {
        tmpoctant[i][j][k]=0;
      }
    }
  }
#ifdef __OLD_PBC
#define binpbc(s,t,u) bin[(s+nbins[0])%nbins[0]][(t+nbins[1])%nbins[1]][(u+nbins[2])%nbins[2]]
#define newbinpbc(s,t,u) newbin[(s+nbins[0])%nbins[0]][(t+nbins[1])%nbins[1]][(u+nbins[2])%nbins[2]]
#define meanbinpbc(s,t,u) meanbin[(s+nbins[0])%nbins[0]][(t+nbins[1])%nbins[1]][(u+nbins[2])%nbins[2]]
#define areabinpbc(s,t,u) areabin[(s+nbins[0])%nbins[0]][(t+nbins[1])%nbins[1]][(u+nbins[2])%nbins[2]]
#define ColIpbc(s,t,u) ColImage[(s+nbins[0])%nbins[0]][(t+nbins[1])%nbins[1]][(u+nbins[2])%nbins[2]]
#else
#define binpbc(s,t,u) pbc_array(bin,s,t,u,nbins,offsets)
#define newbinpbc(s,t,u) pbc_array(newbin,s,t,u,nbins,offsets)
#define meanbinpbc(s,t,u) dpbc_array(meanbin,s,t,u,nbins,offsets)
#define gaussbinpbc(s,t,u) dpbc_array(gaussbin,s,t,u,nbins,offsets)
#define areabinpbc(s,t,u) dpbc_array(areabin,s,t,u,nbins,offsets)
#define ColIpbc(s,t,u) dpbc_array(ColImage,s,t,u,nbins,offsets)
#define gColIpbc(s,t,u) dpbc_array(gColImage,s,t,u,nbins,offsets)
#endif
#define GSSBN(s,t,u) gaussbin[/*(int)(*/s/*/nlc)*/][/*(int)(*/t/*/nlc)*/][/*(int)(*/u/*/nlc)*/]
#define MNBN(s,t,u)   meanbin[/*(int)(*/s/*/nlc)*/][/*(int)(*/t/*/nlc)*/][/*(int)(*/u/*/nlc)*/]
#define RBN(s,t,u)    areabin[/*(int)(*/s/*/nlc)*/][/*(int)(*/t/*/nlc)*/][/*(int)(*/u/*/nlc)*/]
  //loop over bin and setup tmpoctant...
  for (i = 0 ; i < nbins[0] ; i++) {
    for (j = 0 ; j < nbins[1] ; j++) {
      for (k = 0 ; k < nbins[2] ; k++) {
        for (s = 0 ; s < 2 ; s++) { //first set up tmpoctant for every bin...
          for (t = 0 ; t < 2 ; t++) {
            for (u = 0 ; u < 2 ; u++) {
              tmpoctant[s][t][u] = binpbc(i+s,j+t,k+u);
            }
          }
        }
        //then identify tmpoctant
        //MF  here also offsets appeared now = zeromatrix...
        getn3210(oct3210,noct,tmpoctant,zeromatrix); //just a handy way of discriminating, the actual numbers have no meaning
        mybin[i][j][k]=oct3210[3]+oct3210[2]+oct3210[1]+oct3210[0]; //maybe this is not ideal...
        //if this doesn't work, embed the 2x2x2 octant in a 3x3x3 one. This should give more freedom 
        if (mybin[i][j][k]==27) { //case a
          GSSBN(i,j,k)+= 0.5*M_PI; 
          MNBN(i,j,k)+= M_PI*3*chi/8; //THE CURVATURES ARE ALREADY MULTIPLIED WITH THE AREA!!
          RBN(i,j,k)+= 0.75*chi2;
          global_curv[2]+=0.75*chi2;
          global_curv[3]+= M_PI*3*chi/8;
          global_curv[4]+= 0.5*M_PI;
          //printf("a\n");
        }
        else if (mybin[i][j][k]==117) { //case ai
          GSSBN(i,j,k)+= 0.5*M_PI;
          MNBN(i,j,k)+= -M_PI*3*chi/8;
          RBN(i,j,k)+= 0.75*chi2;
          global_curv[2]+=0.75*chi2;
          global_curv[3]+= -M_PI*3*chi/8;
          global_curv[4]+= 0.5*M_PI;
          //printf("ai\n");
        }
        else if (mybin[i][j][k]==45) { //case b
          //GSSBN(i,j,k)+= 0;
          MNBN(i,j,k)+= 0.25*chi*M_PI;
          RBN(i,j,k)+= chi2;
          global_curv[2]+=chi2;
          global_curv[3]+=0.25*chi*M_PI;
          //global_curv[4]+=
          //printf("b\n");
        }
        else if (mybin[i][j][k]==105) { //case bi
          //GSSBN(i,j,k)+= 0;
          MNBN(i,j,k)+= -0.25*chi*M_PI;
          RBN(i,j,k)+= chi2;
          global_curv[2]+=chi2;
          global_curv[3]+= -0.25*chi*M_PI;
          //global_curv[4]+=
          //printf("bi\n");
        }
        else if (mybin[i][j][k]==51) { //case c  
          GSSBN(i,j,k)+= -M_PI;
          MNBN(i,j,k)+= 0.25*chi*M_PI;
          RBN(i,j,k)+= 1.5*chi2;
          global_curv[2]+=1.5*chi2;
          global_curv[3]+=0.25*chi*M_PI;
          global_curv[4]+=-M_PI;
          printf("CASE: c\n");
        }
        else if (mybin[i][j][k]==109) { //case ci 
          GSSBN(i,j,k)+= M_PI;
          MNBN(i,j,k)+= -0.75*chi*M_PI;
          //GSSBN(i,j,k)+= -M_PI;
          //MNBN(i,j,k)+= -0.25*chi*M_PI;
          RBN(i,j,k)+= 1.5*chi2;
          global_curv[2]+=1.5*chi2;
          global_curv[3]+=-0.75*chi*M_PI;
          global_curv[4]+=M_PI;
          printf("CASE: ci\n");
        }
        else if (mybin[i][j][k]==63) { //case e
          GSSBN(i,j,k)+= -0.5*M_PI;
          MNBN(i,j,k)+= 0.125*chi*M_PI;
          RBN(i,j,k)+= 1.25*chi2;
          global_curv[2]+=1.25*chi2;
          global_curv[3]+=0.125*chi*M_PI;
          global_curv[4]+=-0.5*M_PI;
          //printf("e\n");
        }
        else if (mybin[i][j][k]==93) { //case ei
          GSSBN(i,j,k)+= -0.5*M_PI;
          MNBN(i,j,k)+= -0.125*chi*M_PI;
          RBN(i,j,k)+= 1.25*chi2;
          global_curv[2]+=1.25*chi2;
          global_curv[3]+=-0.125*chi*M_PI;
          global_curv[4]+= -0.5*M_PI;
          //printf("ei\n");
        }
        else if (mybin[i][j][k]==69) { //case f
          GSSBN(i,j,k)+= -1.5*M_PI;
          MNBN(i,j,k)+= 0.125*chi*M_PI;
          RBN(i,j,k)+= 1.75*chi2;
          global_curv[2]+=1.75*chi2;
          global_curv[3]+=0.125*chi*M_PI;
          global_curv[4]+=-1.5*M_PI;
          printf("CASE: f\n");
        }
        else if (mybin[i][j][k]==97) { //case fi
          GSSBN(i,j,k)+= 0.5*M_PI;
          MNBN(i,j,k)+= -0.625*chi*M_PI;
          //GSSBN(i,j,k)+= -1.5*M_PI;
          //MNBN(i,j,k)+= -0.125*chi*M_PI;
          RBN(i,j,k)+= 1.75*chi2;
          global_curv[2]+=1.75*chi2;
          global_curv[3]+=-0.625*chi*M_PI;
          global_curv[4]+=0.5*M_PI;
          printf("CASE: fi\n");
        }
        else if (mybin[i][j][k]==75) { //case h
          //GSSBN(i,j,k)+= 0;
          //MNBN(i,j,k)+= 0;
          RBN(i,j,k)+= chi2;
          global_curv[2]+=chi2;
          //global_curv[3]+=
          //global_curv[4]+=
          //printf("h\n");
        }
        else if (mybin[i][j][k]==81) { //case i
          GSSBN(i,j,k)+= -M_PI;
          //MNBN(i,j,k)+= 0;
          RBN(i,j,k)+= 1.5*chi2;
          global_curv[2]+= 1.5*chi2;
          //global_curv[3]+=
          global_curv[4]+=-M_PI;
          //printf("i\n");
        }
        else if (mybin[i][j][k]!=0 && mybin[i][j][k]!=125) {
          printf("CASE: NOISE %d\n",mybin[i][j][k]);
        }
      }
    }
  }
  double scale;
  minMean=-3*M_PI*chi/*EXPERIMENTAL*//8; //MF CHECK THIS! for now leave it for loc-curv.py (event. delete it)
  scale = -1.0/(2*minMean); //make scaling absolute, normalize by largest possible values, not the actual ones
  printf("6PIchi=%18.16f ,scale %18.16f\n",M_PI*6*chi,scale); //MF INFO OUTPUT


  //PREPROCESS IMAGE = DELETE NON-SURFACE PIXELS
  int *** newbin;
  snew(newbin,nbins[0]);
  for (i = 0 ; i<nbins[0] ; i++) {
    snew(newbin[i],nbins[1]);
    for (j = 0 ; j<nbins[1] ; j++) {
      snew(newbin[i][j],nbins[2]); 
    }
  }
  for ( i=0; i<nbins[0]; i++ ) {
    for ( j=0; j<nbins[1]; j++ ) {
      for ( k=0; k<nbins[2]; k++ ) {
        if (bin[i][j][k]) {
          if (!(binpbc(i+1,j,k)&&binpbc(i-1,j,k)&&binpbc(i,j+1,k)&&binpbc(i,j-1,k)&&binpbc(i,j,k+1)&&binpbc(i,j,k-1))){
            newbin[i][j][k]=bin[i][j][k];
          }
        }
      }
    }
  }


  //MAP CURVATURE ON PIXELS
  double *** ColImage, *** nColImage, *** gColImage, *** ngColImage;
  snew(ColImage,nbins[0]);
  snew(nColImage,nbins[0]);
  snew(gColImage,nbins[0]);
  snew(ngColImage,nbins[0]);
  for (i = 0 ; i<nbins[0] ; i++) {
    snew(ColImage[i],nbins[1]);
    snew(nColImage[i],nbins[1]);
    snew(gColImage[i],nbins[1]);
    snew(ngColImage[i],nbins[1]);
    for (j = 0 ; j<nbins[1] ; j++) {
      snew(ColImage[i][j],nbins[2]); 
      snew(nColImage[i][j],nbins[2]); 
      snew(gColImage[i][j],nbins[2]); 
      snew(ngColImage[i][j],nbins[2]); 
    }
  }
  //loop over areabin = VERTICES ----> count adjacent surface PIXELS
  //store in areabin
  for (i=0 ; i<nbins[0] ; i++) {
    for (j=0 ; j<nbins[1] ; j++) {
      for (k=0 ; k<nbins[2] ; k++) {
        if(areabin[i][j][k]>0.0) {
          areabin[i][j][k]=0.2; //safety
          for (s=0 ; s<=1 ; s++) {
            for (t=0 ; t<=1 ; t++) {
              for (u=0 ; u<=1 ; u++) {
                if (newbinpbc(i+s,j+t,k+u)) {
                  areabin[i][j][k]+=1.0; // areabin is still over VERTICES - contains # adjacent surface pixels
                }
              }
            }
          }
        }
      }
    }
  }
  //loop over PIXELS
  minMean=3*M_PI*chi; maxMean=-minMean;
  for (i=0 ; i<nbins[0] ; i++) {
    for (j=0 ; j<nbins[1] ; j++) {
      for (k=0 ; k<nbins[2] ; k++) {
        if(newbin[i][j][k]) {
          for (s=-1 ; s<=0 ; s++) { //loop over adjacent vertices
            for (t=-1 ; t<=0 ; t++) {
              for (u=-1 ; u<=0 ; u++) {
                if (areabinpbc(i+s,j+t,k+u)>0.3) {//=check if this is a surface vertice //MF CIOMIA!!!!
                  ColImage[i][j][k]+=meanbinpbc(i+s,j+t,k+u)/(int)(areabinpbc(i+s,j+t,k+u)); //distribute curv of vertices
                  gColImage[i][j][k]+=gaussbinpbc(i+s,j+t,k+u)/(int)(areabinpbc(i+s,j+t,k+u)); //distribute curv of vertices
                  global_curv[5]+=gaussbinpbc(i+s,j+t,k+u)/(int)(areabinpbc(i+s,j+t,k+u)); //MF DEBUG DELETE THIS
                  global_curv[6]+=meanbinpbc(i+s,j+t,k+u)/(int)(areabinpbc(i+s,j+t,k+u)); //MF DEBUG DELETE THIS
                  //printf("CI %10.9f mb %10.9f ab %d\ni:%d j:%d k:%d, s:%d t:%d u:%d\n",ColImage[i][j][k],meanbinpbc(i+s,j+t,k+u),(int)(areabinpbc(i+s,j+t,k+u)),i,j,k,s,t,u);//MF DEBUG
                }                                             //in equal parts on adjacent pixels (areabin = n.o.pixels,
              }                                              //not area!)
            }
          }
          //get max/min value:
          if (ColImage[i][j][k]>maxMean) maxMean=ColImage[i][j][k];
          if (ColImage[i][j][k]<minMean) minMean=ColImage[i][j][k];
          if (gColImage[i][j][k]>maxGauss) maxGauss=gColImage[i][j][k];
          if (gColImage[i][j][k]<minGauss) minGauss=gColImage[i][j][k];
        }
      }
    }
  }


  //IF YOU WANT AVERAGING, DO IT NOW...
  int counter;
  if (nlc>=1) {
    minMean=0; maxMean=0; minGauss=0; maxGauss=0; //reset min/max
    for (i=0 ; i<nbins[0] ; i++) {
      for (j=0 ; j<nbins[1] ; j++) {
        for (k=0 ; k<nbins[2] ; k++) {
          if (newbin[i][j][k]) {
            counter=0;
            for (s=-nlc ; s<=nlc ; s++) {
              for (t=-nlc ; t<=nlc ; t++) {
                for (u=-nlc ; u<=nlc ; u++) {
                  if (newbinpbc(i+s,j+t,k+u)) {
                    counter++;
                  }
                }
              }
            }
            ColImage[i][j][k]/=counter;
            gColImage[i][j][k]/=counter;
          }
        }
      }
    }
    for (i=0 ; i<nbins[0] ; i++) { //DOES THIS HAVE TO BE A SEPARATE LOOP? SPEED-UP POTENTIAL!!
      for (j=0 ; j<nbins[1] ; j++) {
        for (k=0 ; k<nbins[2] ; k++) {
          if (newbin[i][j][k]) {
            for (s=-nlc ; s<=nlc ; s++) {
              for (t=-nlc ; t<=nlc ; t++) {
                for (u=-nlc ; u<=nlc ; u++) {
                  if (newbinpbc(i+s,j+t,k+u)) {
                    nColImage[i][j][k]+=ColIpbc(i+s,j+t,k+u);
                    ngColImage[i][j][k]+=gColIpbc(i+s,j+t,k+u);
                  }
                }
              }
            }
            if (nColImage[i][j][k]>maxMean) maxMean=nColImage[i][j][k];
            if (nColImage[i][j][k]<minMean) minMean=nColImage[i][j][k]; // no else...
            if (ngColImage[i][j][k]>maxGauss) maxGauss=ngColImage[i][j][k];
            if (ngColImage[i][j][k]<minGauss) minGauss=ngColImage[i][j][k]; // no else...
          }
        }
      }
    }
  }
  //MEAN CURV PIC
  if (minMean==0.0 && maxMean==0) { //no div by 0...
    scale = 1.0;
  } else {
    printf("minMean %18.17f  maxMean %18.17f\n",minMean,maxMean);//MF DEBUG
    if (fabs(minMean)>fabs(maxMean)) {
      scale = fabs(1.0/(2*minMean));
      minMean = - fabs(minMean);
    } else {
      scale = fabs(1.0/(2*maxMean));
      minMean = - fabs(maxMean);
    }
    printf("minMean %18.17f  maxMean %18.17f scale %18.17f\n",minMean,maxMean,scale);//MF DEBUG
  }

  rvec coord,tmpcoord1,tmpcoord2;
  if (nlc>=1) {
    for (i=0 ; i<nbins[0] ; i++) { //ROTATE translate bin into coordinates
      for (j=0 ; j<nbins[1] ; j++) {
        for (k=0 ; k<nbins[2] ; k++) {
          if (newbin[i][j][k]>0) {
            coord[0]=(i+0.5)*chi;
            coord[1]=(j+0.5)*chi;
            coord[2]=(k+0.5)*chi;
            rvec_sub(coord,cntr,tmpcoord1); //THE BACK ROTATION
            mvmul(Z_back,tmpcoord1,tmpcoord2);
            mvmul(Y_back,tmpcoord2,tmpcoord1);
            mvmul(X_back,tmpcoord1,coord);
            if(norm(coord)<distance){ //determine if within radius of interest
              rvec_inc(coord,cntr);
              loc_curv_bin[*loc_curv_bin_length][0]=coord[0];
              loc_curv_bin[*loc_curv_bin_length][1]=coord[1];
              loc_curv_bin[*loc_curv_bin_length][2]=coord[2];
              loc_curv_bin[*loc_curv_bin_length][3]=nColImage[i][j][k];
              loc_curv_bin[*loc_curv_bin_length][4]=ngColImage[i][j][k];
              global_curv[0]+=nColImage[i][j][k];
              global_curv[1]+=ngColImage[i][j][k];
              *loc_curv_bin_length=*loc_curv_bin_length+1;
            }
          }
        }
      }
    }
  } else {
    for (i=0 ; i<nbins[0] ; i++) { //ROTATE translate bin into coordinates
      for (j=0 ; j<nbins[1] ; j++) {
        for (k=0 ; k<nbins[2] ; k++) {
          if (newbin[i][j][k]>0) {
            coord[0]=(i+0.5)*chi;
            coord[1]=(j+0.5)*chi;
            coord[2]=(k+0.5)*chi;
            rvec_sub(coord,cntr,tmpcoord1); //THE BACK ROTATION
            mvmul(Z_back,tmpcoord1,tmpcoord2);
            mvmul(Y_back,tmpcoord2,tmpcoord1);
            mvmul(X_back,tmpcoord1,coord);
            if(norm(coord)<distance){ //determine if within radius of interest
              rvec_inc(coord,cntr);
              loc_curv_bin[*loc_curv_bin_length][0]=coord[0];
              loc_curv_bin[*loc_curv_bin_length][1]=coord[1];
              loc_curv_bin[*loc_curv_bin_length][2]=coord[2];
              loc_curv_bin[*loc_curv_bin_length][3]=ColImage[i][j][k];
              loc_curv_bin[*loc_curv_bin_length][4]=gColImage[i][j][k];
              global_curv[0]+=ColImage[i][j][k];
              global_curv[1]+=gColImage[i][j][k];
              *loc_curv_bin_length=*loc_curv_bin_length+1;
            }
          }
        }
      }
    }
  }

  //GAUSS CURV PIC
  if (minGauss==0.0 && maxGauss==0) { //no div by 0...
    scale = 1.0;
  } else {
    printf("minGauss %18.17f  maxGauss %18.17f\n",minGauss,maxGauss);//MF DEBUG
    if (fabs(minGauss)>fabs(maxGauss)) {
      scale = fabs(1.0/(2*minGauss));
      minGauss = - fabs(minGauss);
    } else {
      scale = fabs(1.0/(2*maxGauss));
      minGauss = - fabs(maxGauss);
    }
    printf("minGauss %18.17f  maxGauss %18.17f scale %18.17f\n",minGauss,maxGauss,scale);//MF DEBUG
  }

  if (nlc>=1) {
    //write_color_image(fp5,newbin,nbins,chi/*dimension*/,ngColImage,scale,minGauss);
  } else {
    //write_color_image(fp5,newbin,nbins,chi/*dimension*/,gColImage,scale,minGauss);
  }



  freebin(&tmpoctant,noct);
  freebin(&ColImage,nbins); //not completely nice - freebin expects a pointer to integers, bus sfree doesn't mind?
  freebin(&nColImage,nbins); //not completely nice - freebin expects a pointer to integers, bus sfree doesn't mind?
  freebin(&gColImage,nbins); //not completely nice - freebin expects a pointer to integers, bus sfree doesn't mind?
  freebin(&ngColImage,nbins); //not completely nice - freebin expects a pointer to integers, bus sfree doesn't mind?
  freebin(&newbin,nbins);
  sfree(oct3210);
  for (i = 0 ; i < 3 ; i++)
    sfree(zeromatrix[i]);
  sfree(zeromatrix);
  return mybin;
}

int*** binning(t_trxframe frame,atom_id **index,int*nindex,int np,real dim,ivec* nbins,int *npts,real sigma,ivec** ndx,int nlc, bool LV, int thresh, int cs, int*** offsets, matrix X_rot, matrix Y_rot, matrix Z_rot) {
  if (np<1||np>2) {
    printf("Error: np has to be 1 or 2!");
    exit(9);
  }
  int temp,nn,phase,i,j,k,plusminus;
  plusminus = 1;
  ivec tmp;
  for (i = 0 ; i<3 ; i++) { //MFDEBUG
    for (j = 0 ; j<3 ; j++) {
      printf("box[%d][%d]=%8.5f\n",i,j,frame.box[i][j]);
    }
  } //MFDEBUG
  snew(*offsets,3);
  for (i = 0 ; i<3 ; i++) {
    snew((*offsets)[i],3);
  }
  //determine number of bins
  for (i=0 ; i<DIM ; i++) {
    (*nbins)[i]=(int)(frame.box[i][i]/dim);
    if ((*nbins)[i]*dim>frame.box[i][i]) { //MFDEBUG took out = 
      printf("Error: (*nbins)[%d]*dim=%8.5f > frame.box[%d]=%8.5f\n",i,(*nbins)[i]*dim,i,frame.box[i][i]);
    }
    fprintf(stderr,"(*nbins) %d i %d natoms %d\n",(*nbins)[i],i,frame.natoms); //MF DEBUG
  }
  //allocate bin
  int*** mybin;
  rvec* sphere;
  snew(mybin,(*nbins)[0]);
  for (i = 0 ; i<(*nbins)[0] ; i++) {
    snew(mybin[i],(*nbins)[1]);
    for (j = 0 ; j<(*nbins)[1] ; j++) {
      snew(mybin[i][j],(*nbins)[2]); 
    }
  }
  //determine offsets
  for (i=0 ; i<3 ; i++) {
    for (j=0 ; j<i ; j++) {
      (*offsets)[i][j] = (int)(frame.box[i][j]/dim);
    }
  }


  put_atoms_in_box_and_make_offset_multpl_dim(frame.box,frame.natoms,frame.x,dim);

  rvec cntr;
  cntr[0]=0.5*frame.box[0][0];cntr[1]=0.5*frame.box[1][1];cntr[2]=0.5*frame.box[2][2];

  //loop over atoms in index
  for (phase = 0 ; phase < np ; phase++) {  
    for (nn=0 ; nn < nindex[phase] ; nn++) {
      //expand atom to sphere (not one coordinate, but npts points)
      /*ROTATE IN THIS STEP*/
      sphere = probe_spheres_while_rotating(sigma, npts, frame.x[index[phase][nn]],frame.box,dim,X_rot,Y_rot,Z_rot,cntr);
      if (sphere[0][0]>-9999.0) {
        for (j = 0 ; j < *npts ; j++) {
          for (i = 0 ; i < DIM ; i++) {
            tmp[i]=(int)(sphere[j][i]/frame.box[i][i]*(*nbins)[i]); //SPEED UP POTENTIAL!!!
            if (tmp[i]==(*nbins)[i]) { //MF SAFETY MEASURE
              fprintf(stderr,"tmp: %d, box: %d\nreducing tmp by 1!\n",tmp[i],(*nbins)[i]);
              tmp[i]--;
            }
          }
          mybin[tmp[0]][tmp[1]][tmp[2]]+=plusminus;
        }
      }
      sfree(sphere);
    }
    plusminus=-1; //we do the above once for the positive phase. if the negative phase is treated explicitely
		  //(np = 2) we do this ONE more time, this time subtracting...
  }
  //convert bin to 0 or 1 - format expected in getn3210
  //and create ndx of background... which is needed for the cluster filtering
  snew(*ndx,(*nbins)[0]*(*nbins)[1]*(*nbins)[2]+1);//ndx: 0 length([0][0]), 1,2,3... indices of nth element
  (*ndx)[0][0]=1;
  
  for (i = 0 ; i<(*nbins)[0] ; i++) {
    for (j = 0 ; j<(*nbins)[1] ; j++) {
      for (k = 0 ; k<(*nbins)[2] ; k++) {
        if (mybin[i][j][k]>thresh) { //FOREGROUND
          mybin[i][j][k]=1;
          //DO I ALSO NEED AN NDX OF THE FOREGROUND?? MF IMAGEMANIPULATION...
        } else {                     //BACKGROUND
          if (cs>0) {
            (*ndx)[(*ndx)[0][0]][0]=i;
            (*ndx)[(*ndx)[0][0]][1]=j;
            (*ndx)[(*ndx)[0][0]][2]=k;
            mybin[i][j][k]=-(*ndx)[0][0]+1; //+1 to number the background from [0-#bgd[
            (*ndx)[0][0]++;
          }
          else {
            mybin[i][j][k]=0;
          }
        }
      }
    }
  }
  return mybin;
}

void freebin(int**** bin, ivec nbins) {
  int i,j;
  for (i = 0 ; i<nbins[0] ; i++) {
    for (j = 0 ; j<nbins[1] ; j++) {
      sfree((*bin)[i][j]); 
    }
    sfree((*bin)[i]);
  }
  sfree(*bin);
}

void getn3210(int* n3210,ivec nbins,int*** GRID,int** offsets) {//computes number of active cubes, faces, edges and vertices
                                                   
  int xx,yy,zz,aa,bb,cc,i,j;
  //allocate dual Grid
  int *** DGRID;
  snew(DGRID,nbins[0]);
  for (i = 0 ; i<nbins[0] ; i++) {
    snew(DGRID[i],nbins[1]);
    for (j = 0 ; j<nbins[1] ; j++) {
      snew(DGRID[i][j],nbins[2]); 
    }
  }
  //set n3210 to zero
  for (i = 0 ; i < 4 ; i++)
    n3210[i] = 0;
  //loop over Grid (uses definitions! idgrd contains modulus --> PBC)
  for (xx = 0 ; xx < nbins[0] ; xx++) {
    for (yy = 0 ; yy < nbins[1] ; yy++) {
      for (zz = 0 ; zz < nbins[2] ; zz++) {
        if (grd(xx,yy,zz)==1) {
          dgrd(xx,yy,zz)=1;
          n3210[3] += 1;
          for (aa = -1 ; aa <= 1 ; aa += 2) {
            n3210[2] += idgrd(xx+aa,yy,zz) + idgrd(xx,yy+aa,zz) + idgrd(xx,yy,zz+aa);
            for (bb = -1 ; bb <= 1 ; bb += 2) {
              n3210[1] += idgrd(xx+aa,yy,zz)*idgrd(xx+aa,yy+bb,zz)*idgrd(xx,yy+bb,zz)
                        + idgrd(xx+aa,yy,zz)*idgrd(xx+aa,yy,zz+bb)*idgrd(xx,yy,zz+bb)
                        + idgrd(xx,yy+aa,zz)*idgrd(xx,yy+aa,zz+bb)*idgrd(xx,yy,zz+bb);
              for (cc = -1 ; cc <= 1 ; cc += 2) {
                n3210[0] += idgrd(xx+aa,yy,zz)*idgrd(xx+aa,yy+bb,zz)*idgrd(xx+aa,yy,zz+cc)
                          * idgrd(xx+aa,yy+bb,zz+cc)*idgrd(xx,yy+bb,zz)
                          * idgrd(xx,yy+bb,zz+cc)*idgrd(xx,yy,zz+cc);
              }
            }
          }
        }
      }
    }
  }
  freebin(&DGRID,nbins);
}

int eul3d(int* nn) {
  printf("3: %d 2: %d 1: %d 0: %d\n",nn[3],nn[2],nn[1],nn[0]);
  return -nn[3]+nn[2]-nn[1]+nn[0];
}
real vol3d(int* nn, real edgelength) {
  return nn[3]*pow(edgelength,3);
}
real sur3d(int* nn, real edgelength) {
  return (-6*nn[3]+2*nn[2])*pow(edgelength,2);
}
int twoB3d(int* nn, real edgelength) {
  return (3*nn[3]-2*nn[2]+nn[1])*edgelength;
}

bool neighbour_bool(ivec ind1, ivec ind2, ivec dim) {
  int i,nb;
  nb=0;
  for (i = 0 ; i < DIM ; i++) {
    if (ind1[i] == ind2[i])
      nb+=10;
    else if (abs(ind1[i]-ind2[i]) == 1 || abs(ind1[i]-ind2[i]) == dim[i]-1)
      nb+=1;
    else 
      break;
  }
  return (nb == 21);
}

bool ImgMan(ivec* coord, int**** image, ivec nbins, int** offsets) {
//ELIMINATE BUMPS/DENTS, CHANGE coord TO COORDINATES OF SINGLE (SIMILAR) NEIGHBOUR
  int val, nval, i, j, k, count;
  i = (*coord)[0]; j = (*coord)[1]; k = (*coord)[2];
  val = pbc_array(*image,i,j,k,nbins,offsets);
  nval = labs(val-1);
  count=0;
  if(pbc_array(*image,i+1,j,k,nbins,offsets)==val) {
    count++;
    (*coord)[0]=i+1;
  }
  if(pbc_array(*image,i-1,j,k,nbins,offsets)==val) {
    count++;
    if (count==1) { (*coord)[0]=i-1; }
    else { (*coord)[0]=-23; return FALSE; } //leave this line in all cases... (except first)
  }
  if(pbc_array(*image,i,j+1,k,nbins,offsets)==val) {
    count++;
    if (count==1) { (*coord)[1]=j+1; }
    else { (*coord)[0]=-23; return FALSE; } //leave this line in all cases... (except first)
  }
  if(pbc_array(*image,i,j-1,k,nbins,offsets)==val) {
    count++;
    if (count==1) { (*coord)[1]=j-1; }
    else { (*coord)[0]=-23; return FALSE; } //leave this line in all cases... (except first)
  }
  if(pbc_array(*image,i,j,k+1,nbins,offsets)==val) {
    count++;
    if (count==1) { (*coord)[2]=k+1; }
    else { (*coord)[0]=-23; return FALSE; } //leave this line in all cases... (except first)
  }
  if(pbc_array(*image,i,j,k-1,nbins,offsets)==val) {
    count++;
    if (count==1) { (*coord)[2]=k-1; }
    else { (*coord)[0]=-23; return FALSE; } //leave this line in all cases... (except first)
  }
  if (count==0) {(*coord)[0]=-23; return FALSE; }
  else {
    pbc_array_w(image,i,j,k,nbins,offsets,nval); //invert pixel
    return TRUE;
  }
}

void cluster_filter(int**** image, ivec nbins, ivec* ndx, int size, int** offsets) {
  int i,j,k,a,b,c;
  int** cluster; //0 will be the offset of the last member + 1, that is the length of the array
  //MAYBE ALLOCATE CLUSTERS ONLY FULLY, IF THEY ARE A "EATER"??? MEMORY MF
  snew(cluster,ndx[0][0]);
  for (i = 0 ; i < ndx[0][0] ; i++) {
    snew(cluster[i],/*ndx[0][0]-i+2*/3); //do not yet allocate...
    cluster[i][0]=1;
    cluster[i][1]=i;
    cluster[i][2]=-22;
  }
#ifdef __OLD_PBC
#define img(s,t,u) (*image)[(s+nbins[0])%nbins[0]][(t+nbins[1])%nbins[1]][(u+nbins[2])%nbins[2]]
#else
#define img(s,t,u) pbc_array((*image),s,t,u,nbins,offsets)
#define wimg(s,t,u,v) pbc_array_w((image),(s),(t),(u),nbins,offsets,(v))
#endif


  //find clusters...
  int clusternumber;
  //clusternumber = ndx[0][0];
  clusternumber = ndx[0][0]-1; //MF NEWCHANGE added -1!!! ndx[0][0] = #x*#y*#z+1 [7.2.2008]
        fprintf(stderr,"initial clusternumber %d\n",clusternumber);
  for (i = 0 ; i < ndx[0][0]-1 ; i++) { //MF NEWCHANGE added -1
    (*image)[ndx[i+1][0]][ndx[i+1][1]][ndx[i+1][2]]=0;//temp. stored the backgrnd index there, need 0 for getn3210
                                                      //this one is a eater...
    for (a = 1 ; a <= cluster[i][0] ; a++) { //necessary if cluster eats another one 
      if (cluster[i][2]==-22) { //potential eater <== cluster[i][0] != 0
        fprintf(stderr,"got a -22! MF22\n");
        sfree(cluster[i]);              //free and
        snew(cluster[i],ndx[0][0]-i+2); //allocate again (this time for worst case)
        cluster[i][0]=1;                //you might want to use realloc instead...
        cluster[i][1]=i;
      }

      //ELIMINATE LOOP INTO SOMETHING EXPLICIT USING ONLY NEIGHBOURING INDICES!!!
        if (img(ndx[cluster[i][a]+1][0]+1,ndx[cluster[i][a]+1][1],ndx[cluster[i][a]+1][2])<0) {
                                                        //eaten ones/ex-eaters: 0! uneaten ones: negative!
          //eat it!! - increase length of i by one and copy its ndx into i
          cluster[i][0]++;
          cluster[i][cluster[i][0]]=cluster[-img(ndx[cluster[i][a]+1][0]+1,ndx[cluster[i][a]+1][1],ndx[cluster[i][a]+1][2])][1];
          //and eliminate it
          cluster[-img(ndx[cluster[i][a]+1][0]+1,ndx[cluster[i][a]+1][1],ndx[cluster[i][a]+1][2])][0]=0;
          //and make a note of eater in it
          clusternumber--;
#ifdef __OLD_PBC
          img(ndx[cluster[i][a]+1][0]+1,ndx[cluster[i][a]+1][1],ndx[cluster[i][a]+1][2])=0; //mark as eaten
#else
          wimg(ndx[cluster[i][a]+1][0]+1,ndx[cluster[i][a]+1][1],ndx[cluster[i][a]+1][2],0); //mark as eaten
#endif
        }
        if (img(ndx[cluster[i][a]+1][0]-1,ndx[cluster[i][a]+1][1],ndx[cluster[i][a]+1][2])<0) {
          //eat it!! - increase length of i by one and copy its ndx into i
          cluster[i][0]++;
          cluster[i][cluster[i][0]]=cluster[-img(ndx[cluster[i][a]+1][0]-1,ndx[cluster[i][a]+1][1],ndx[cluster[i][a]+1][2])][1];
          //and eliminate it
          cluster[-img(ndx[cluster[i][a]+1][0]-1,ndx[cluster[i][a]+1][1],ndx[cluster[i][a]+1][2])][0]=0;
          clusternumber--;
#ifdef __OLD_PBC
          img(ndx[cluster[i][a]+1][0]-1,ndx[cluster[i][a]+1][1],ndx[cluster[i][a]+1][2])=0; //mark as eaten
#else
          wimg(ndx[cluster[i][a]+1][0]-1,ndx[cluster[i][a]+1][1],ndx[cluster[i][a]+1][2],0); //mark as eaten
#endif
        }
        if (img(ndx[cluster[i][a]+1][0],ndx[cluster[i][a]+1][1]+1,ndx[cluster[i][a]+1][2])<0) {
          //eat it!! - increase length of i by one and copy its ndx into i
          cluster[i][0]++;
          cluster[i][cluster[i][0]]=cluster[-img(ndx[cluster[i][a]+1][0],ndx[cluster[i][a]+1][1]+1,ndx[cluster[i][a]+1][2])][1];
          //and eliminate it
          cluster[-img(ndx[cluster[i][a]+1][0],ndx[cluster[i][a]+1][1]+1,ndx[cluster[i][a]+1][2])][0]=0;
          clusternumber--;
#ifdef __OLD_PBC
          img(ndx[cluster[i][a]+1][0],ndx[cluster[i][a]+1][1]+1,ndx[cluster[i][a]+1][2])=0; //mark as eaten
#else
          wimg(ndx[cluster[i][a]+1][0],ndx[cluster[i][a]+1][1]+1,ndx[cluster[i][a]+1][2],0); //mark as eaten
#endif
        }
        if (img(ndx[cluster[i][a]+1][0],ndx[cluster[i][a]+1][1]-1,ndx[cluster[i][a]+1][2])<0) {
          //eat it!! - increase length of i by one and copy its ndx into i
          cluster[i][0]++;
          cluster[i][cluster[i][0]]=cluster[-img(ndx[cluster[i][a]+1][0],ndx[cluster[i][a]+1][1]-1,ndx[cluster[i][a]+1][2])][1];
          //and eliminate it
          cluster[-img(ndx[cluster[i][a]+1][0],ndx[cluster[i][a]+1][1]-1,ndx[cluster[i][a]+1][2])][0]=0;
          clusternumber--;
#ifdef __OLD_PBC
          img(ndx[cluster[i][a]+1][0],ndx[cluster[i][a]+1][1]-1,ndx[cluster[i][a]+1][2])=0; //mark as eaten
#else
          wimg(ndx[cluster[i][a]+1][0],ndx[cluster[i][a]+1][1]-1,ndx[cluster[i][a]+1][2],0); //mark as eaten
#endif
        }
        if (img(ndx[cluster[i][a]+1][0],ndx[cluster[i][a]+1][1],ndx[cluster[i][a]+1][2]+1)<0) {
          //eat it!! - increase length of i by one and copy its ndx into i
          cluster[i][0]++;
          cluster[i][cluster[i][0]]=cluster[-img(ndx[cluster[i][a]+1][0],ndx[cluster[i][a]+1][1],ndx[cluster[i][a]+1][2]+1)][1];
          //and eliminate it
          cluster[-img(ndx[cluster[i][a]+1][0],ndx[cluster[i][a]+1][1],ndx[cluster[i][a]+1][2]+1)][0]=0;
          clusternumber--;
#ifdef __OLD_PBC
          img(ndx[cluster[i][a]+1][0],ndx[cluster[i][a]+1][1],ndx[cluster[i][a]+1][2]+1)=0; //mark as eaten
#else
          wimg(ndx[cluster[i][a]+1][0],ndx[cluster[i][a]+1][1],ndx[cluster[i][a]+1][2]+1,0); //mark as eaten
#endif
        }
        if (img(ndx[cluster[i][a]+1][0],ndx[cluster[i][a]+1][1],ndx[cluster[i][a]+1][2]-1)<0) {
          //eat it!! - increase length of i by one and copy its ndx into i
          cluster[i][0]++;
          cluster[i][cluster[i][0]]=cluster[-img(ndx[cluster[i][a]+1][0],ndx[cluster[i][a]+1][1],ndx[cluster[i][a]+1][2]-1)][1];
          //and eliminate it
          cluster[-img(ndx[cluster[i][a]+1][0],ndx[cluster[i][a]+1][1],ndx[cluster[i][a]+1][2]-1)][0]=0;
          clusternumber--;
#ifdef __OLD_PBC
          img(ndx[cluster[i][a]+1][0],ndx[cluster[i][a]+1][1],ndx[cluster[i][a]+1][2]-1)=0; //mark as eaten
#else
          wimg(ndx[cluster[i][a]+1][0],ndx[cluster[i][a]+1][1],ndx[cluster[i][a]+1][2]-1,0); //mark as eaten
#endif
        }

    }
  }

          printf("clusternumber = %d ndx[0][0] = %d\n",clusternumber,ndx[0][0]);
  //...and filter clusters smaller than "size"
  for (i = 0 ; i < ndx[0][0]-1 ; i++) { //MF  because ndx[0][0] starts at 1 and is increased by one per bgrd-pixel
                                        //the index of the last entry in the actual list (starts at 0) is ndx[0][0]-2
    if  (cluster[i][0]>0) printf("cluster[%d][0] = %d\n",i,cluster[i][0]);
    if (cluster[i][0]>0 && cluster[i][0]<=size) {
      for (j = 1 ; j <= cluster[i][0] ; j++) {
        (*image)[ndx[cluster[i][j]+1][0]][ndx[cluster[i][j]+1][1]][ndx[cluster[i][j]+1][2]]=1;
          printf("still filtering x:%d y:%d z:%d\n",ndx[cluster[i][j]+1][0],ndx[cluster[i][j]+1][1],ndx[cluster[i][j]+1][2]); //MF DEBUG
      }
    }
  }
  for (i = 0 ; i < ndx[0][0] ; i++) {
    sfree(cluster[i]);
  }
  sfree(cluster);
}

int gmx_mialoc(int argc,char *argv[])
{
  static char *desc[] = {
    "This is a tool for the morphological analysis of ",
    "binary mixtures. Use this executable to obtain the ",
    "averaged local values of the Gaussian and mean ",
    "curvature. ",
    "Note that this version only generates output for a ",
    "single snapshot and that the area of interest must be ",
    "centered in the box. To handle the orientational averaging, ",
    "this executable ignores periodic boundary conditions. ",
    "To obtain the global values of the Minkowski functionals, ",
    "use \"g_mia\" instead.",
  };
  
  static real bindim=0.5;
  static real radius=0.235;
  static int np=1;
  static int nlc=2;
  static int nlc2=2;
  static bool LV=1;
  static int cs=1;
  static int npts=27;
  static int thresh=0;
  static int thresh2=1;
  static int xr=6;
  static int yr=6;
  static int zr=6;
  static real dphi=0.89759701;
  static real depsilon=0.89759701;
  static real dtheta=0.89759701;
  static real dis_fac=0.7;
  int i,j,k,s,t,u,counter; //counters for many things, leave them!
  int *** bin;   //grid: used first for counting atom-dots in the cell and then converted to b/w PIC
  int *** locID;   //grid: stores identifier for local curvatures
  ivec* grdndx; //grdndx = NULL;
  bool VISIBILI = FALSE;
  real phi,epsilon,theta,dANGLE,tmp,distance;
  real **loc_curv_bin=NULL;
  int loc_curv_bin_length;
  real min_g,max_g,min_m,max_m,scale;
  real*** gbin;
  real*** ngbin;
  real*** mbin;
  real*** nmbin;
  int*** cbin;

  /* Extra arguments - but note how you always get the begin/end
   * options when running the program, without mentioning them here!
   */
  
  t_pargs pa[] = {
    //{ "-res1", FALSE, etSTR,  {&res_name1},"Name of residue that specifies the negative phase"      },
    //{ "-res2", FALSE, etSTR,  {&res_name2},"Name of residue that specifies the negative phase"      },
    //{ "-res3", FALSE, etSTR,  {&res_name3},"Name of residue that specifies the negative phase"      },
    //{ "-nres", FALSE, etINT,  {&nres},     "Number of different residues in negative phase (max. 3)"},
    { "-np", FALSE, etINT,  {&np},         "Specifies wether to treat the negative phase explicitely (2) or not (1)"},
    { "-locval", FALSE, etBOOL,  {&LV},    "Compute local curvatures (1) or not (0)?"},
    { "-ar1", FALSE, etINT,  {&nlc},       "Averaging is performed over neighbouring surface cells up to ar1 cells away"},
    { "-ar2", FALSE, etINT,  {&nlc2},     "Averaging after the rotations is performed over neighbouring surface cells up to ar2 cells away"},
    { "-cs"  , FALSE, etINT,  {&cs},       "Maximum number of clustered background grids that are filtered"},
    { "-npts", FALSE, etINT,  {&npts},     "Number of points used for filling of sphere"},
    { "-thresh1", FALSE, etINT,  {&thresh}, "Threshold: minimum number of points required to count a grid cell as positive"},
    { "-thresh2", FALSE, etINT, {&thresh2},"Threshold: minimum number of points required to count a grid cell as positive"},
    { "-nx", FALSE, etINT, {&xr},          "Number of rotations around x"},
    { "-ny", FALSE, etINT, {&yr},          "Number of rotations around y"},
    { "-nz", FALSE, etINT, {&zr},          "Number of rotations around z"},
    { "-depsilon"  , FALSE, etREAL, {&dphi},   "angle-step of rotations around x"},
    { "-dphi"  , FALSE, etREAL, {&depsilon},   "angle-step of rotations around y"},
    { "-dtheta"  , FALSE, etREAL, {&dtheta},   "angle-step of rotations around z"},
    { "-sr"  , FALSE, etREAL, {&radius},   "Radius of spheres used for beads in nm"},
    { "-dim" , FALSE, etREAL, {&bindim},   "Edge length of cubes for grid in nm "},
    { "-dr", FALSE, etREAL, {&dis_fac},"Factor with which half of the minimum box dimension is multiplied to determine the maximum distance from the box center at which grid cells are assumed interesting"}
  };
  
  //t_topology top;
  char       title[STRLEN];
  t_trxframe fr;
  matrix     box;
  matrix     X_rot,Y_rot,Z_rot,X_back,Y_back,Z_back;
  int        status;
  int        flags = TRX_READ_X;
  int        atnr;  //counter
  ivec       nbins; //dimension of bin/GRID/DGRID
  int**      offsets; //offsets of periodic image in triclinic boxes
  int*       n3210; //number of cubes, faces, edges, vertices
  snew(n3210,4);
  t_atoms atoms;
  atom_id    **index;    //numbers of index file
  int        *nindex;    //length of groups in index
  char       **indexgrp; //name of index groups - not used, but required for formal reasons...
  ivec       tmpvec;
  real       *global_curv;
  snew(global_curv,7);

  t_filenm fnm[] = {
    //{ efTPS,  NULL,  NULL, ffREAD },   /* this is for the topology     */
    { efTRX, "-f", NULL, ffREAD },     /* and this for the trajectory  */
    { efNDX, "-n", NULL, ffREAD },     /* and this for the index  */
    { efXVG,"-o","mia",ffWRITE }       /* xvgr output file 	       */
  };
  
#define NFILE asize(fnm)

  CopyRight(stderr,argv[0]);

  /* This is the routine responsible for adding default options,
   * calling the X/motif interface, etc. */
  parse_common_args(&argc,argv,PCA_CAN_TIME | PCA_CAN_VIEW,
		    NFILE,fnm,asize(pa),pa,asize(desc),desc,0,NULL);

  //printf("sizeof: int %d real %d float %d double %d\n",sizeof(int),sizeof(real),sizeof(float),sizeof(double));
  if (thresh2<1) {printf("thresh2 has to be at least 1!\n");exit(4234);}


  FILE* fp;
  fp = xvgropen(opt2fn("-o",NFILE,fnm),"MIA","Time [ps]","Minkowski Functionals");

  
  /* The first time we read data is a little special */
  read_first_frame(&status,ftp2fn(efTRX,NFILE,fnm),&fr,flags);

  snew(nindex,np);
  snew(index,np);
  snew(indexgrp,np);
  get_index(&atoms,ftp2fn_null(efNDX,NFILE,fnm),np,nindex,index,indexgrp);
   


  /* This is the main loop over frames */
  do {
    /* coordinates are available in the vector fr.x
     * you can find this and all other structures in
     * the types directory under the gromacs include dir.
     * Note how flags determines wheter to read x/v/f!
     */
    //printf("Coordinates at t=%8.3f : %8.5f %8.5f %8.5f\n",fr.time,fr.x[1][XX],fr.x[1][YY],fr.x[1][ZZ]);
    for (i=0;i<7;i++)
      global_curv[i]=0;

    tmp=fr.box[0][0]; //DETERMINE distance from center wherin the grid cells are used for the local curvature
                  fprintf(stderr,"tmp: %8.5f frame: %8.5f\n",tmp,fr.box[0][0]);
    for(i=1;i<3;i++){
      if (fr.box[i][i]<tmp) {
        tmp=fr.box[i][i];
      }
                  fprintf(stderr,"tmp: %8.5f frame: %8.5f\n",tmp,fr.box[i][i]);
    }
    distance=dis_fac*tmp*0.5;
                  fprintf(stderr,"distance: %8.5f\n",distance);


    /*ROTATE###########################################################################*/
    //loop over rotations
    //dANGLE  = (real)(M_PI/10.0); //5 steps from 0 to 0.5 PI
    //dANGLE  = (real)(M_PI/14.0); //5 steps from 0 to 0.5 PI
    for(s=0;s<=xr;s++) { //phi
     for(t=0;t<=yr;t++) { //epsilon
      for(u=0;u<=zr;u++) { //theta
        phi=s*dphi; epsilon=t*depsilon; theta=u*dtheta;
        fprintf(stderr,"phi: %8.5f epsilon: %8.5f theta: %8.5f\n",phi,epsilon,theta);
        gen_rot_matrices(X_rot,Y_rot,Z_rot,phi,epsilon,theta);    
        m_inv(X_rot,X_back);
        m_inv(Y_rot,Y_back);
        m_inv(Z_rot,Z_back);

        //MF DEBUG TEST MATRICES
        /*
        rvec mtv1,mtv2,myvec; myvec[0]=7; myvec[1]=6; myvec[2]=5; 
        mtv2[0]=5; mtv2[1]=5; mtv2[2]=5;
        fprintf(stderr,"MMvec1:%f/%f/%f\n",myvec[0],myvec[1],myvec[2]); //MF DEBUG
        rvec_sub(myvec,mtv2,mtv1);
        fprintf(stderr,"MMvec2:%f/%f/%f\n",mtv1[0],mtv1[1],mtv1[2]); //MF DEBUG
        mvmul(X_rot,mtv1,myvec);
        fprintf(stderr,"MMvec3:%f/%f/%f\n",myvec[0],myvec[1],myvec[2]); //MF DEBUG
        mvmul(Z_rot,myvec,mtv1);
        fprintf(stderr,"MMvec4:%f/%f/%f\n",mtv1[0],mtv1[1],mtv1[2]); //MF DEBUG
        mvmul(Z_back,mtv1,myvec);
        fprintf(stderr,"MMvec5:%f/%f/%f\n",myvec[0],myvec[1],myvec[2]); //MF DEBUG
        mvmul(X_back,myvec,mtv1);
        fprintf(stderr,"MMvec6:%f/%f/%f\n",mtv1[0],mtv1[1],mtv1[2]); //MF DEBUG
        rvec_inc(mtv1,mtv2);
        fprintf(stderr,"MMvec7:%f/%f/%f\n",mtv1[0],mtv1[1],mtv1[2]); //MF DEBUG
        */


        /*rotate everything within binning!!!################################################*/
       
        //NEW AND BETTER ALL IN ONE BINNING FUNCTION!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
        bin = binning(fr, index, nindex, np, bindim, &nbins, &npts, radius, &grdndx,nlc,LV,thresh,cs,&offsets,X_rot,Y_rot,Z_rot);
       
              fprintf(stderr,"done binning\n");
              printf("for cbins - gridcells:%d background:%d diff:%d\n",nbins[0]*nbins[1]*nbins[2],grdndx[0][0],nbins[0]*nbins[1]*nbins[2]-grdndx[0][0]);
        
        //apply Cluster Filter
        if (cs>0) {  
          cluster_filter(&bin, nbins, grdndx, cs, offsets);
              fprintf(stderr,"done filtering\n");
        }
       
        //IMAGE MANIPULATION
        for (i = 0 ; i<nbins[0] ; i++) {
          for (j = 0 ; j<nbins[1] ; j++) {
            for (k = 0 ; k<nbins[2] ; k++) {
              tmpvec[0]=i;tmpvec[1]=j;tmpvec[2]=k;
              while(tmpvec[0]!=-23) {
                ImgMan(&tmpvec, &bin, nbins, offsets);
              }
            }
          }
        }
        fprintf(stderr,"done IMANing\n");
       
       
        //NOW FOR THE LOCAL CURVATURES!!!!!!!!!!!!!
        if (LV) {
          if(loc_curv_bin == NULL) {
            fprintf(stderr,"loc_curv_bin == NULL\n");
            //j=nbins[0]*nbins[1]*nbins[2]*25+1;
            j=nbins[0]*nbins[1]*nbins[2]*(xr+1)*(yr+1)*(zr+1)+1;
            snew(loc_curv_bin,j);//MF ATTENTION - THE NUMBER OF ROTATIONS IS HARD CODED. SHOULD THAT CHANGE - TROUBLE!!!
            for(i=0;i<j;i++) {
              snew(loc_curv_bin[i],5); //3coords+mean+curv
            }
            loc_curv_bin_length=0;
          }
          locID = get_locID(bin,nbins,nlc,bindim,offsets,loc_curv_bin,&loc_curv_bin_length,X_back,Y_back,Z_back,distance,global_curv);
          fprintf(stderr,"done get_locID\n");
          /*ROTATE somehow extract GAUSS and MEANBin from function
          retranslate in coordinates ROTATING BACK TO THE ORIGINAL!!!!
          MAYBE ALLOCATE ARRAY AT BEGINNING OF #Rot*natoms? storing the 3 coords and the 2 curvs
          LOOP over that array, BINNING, adding the curvs and keeping tracks of the points in that bin (NO SPHERE PROBING!!!).
          APPLY MINIMUM NUMBER OF POINTS to count the bin as positive and average positive bins by dividing by the points.
       
          ALSO: DELETE ALL POINTS AT A DISTANCE GREATER THAN R_INTEREST FROM CENTER! this takes care of meaningless 
          curvatures at the boundaries*/
       
          freebin(&locID,nbins);
        }
    

        for (i=0 ; i<3 ; i++) {
          sfree(offsets[i]);
        }
        sfree(offsets);
        freebin(&bin,nbins);
        sfree(grdndx);
      } //theta-loop
     } //epsilon-loop
    } //phi-loop
    fprintf(stderr,"done with rotation loops\n");
    
    //ROTATE - make pictures here:
    //MAP BACK ONTO GRID, while counting the contributors
    //MEDIAN FILTERING INSTEAD OF AVERAGING???? - doesn't work easily, since no contribution from neg voxels...
    //IF NO MEDIAN FILTERING - USE THRESHOLD TO DISCARD VOXELS WITH TOO FEW CONTRIBUTORS!
    //HOW ABOUT MEDIAN FILTERING AFTER DISCARDING THE VOXELS WITH TOO LITTLE CONTRIBUTORS?
    //for now, stick with standard averaging...
    if (LV) {
      min_g=0.0;max_g=0.0;min_m=0.0;max_m=0.0;
      snew(cbin,nbins[0]);
      snew(mbin,nbins[0]);
      snew(gbin,nbins[0]);
      snew(nmbin,nbins[0]);
      snew(ngbin,nbins[0]);
      for (i = 0 ; i<nbins[0] ; i++) {
        snew(cbin[i],nbins[1]);
        snew(mbin[i],nbins[1]);
        snew(gbin[i],nbins[1]);
        snew(nmbin[i],nbins[1]);
        snew(ngbin[i],nbins[1]);
        for (j = 0 ; j<nbins[1] ; j++) {
          snew(cbin[i][j],nbins[2]); 
          snew(mbin[i][j],nbins[2]); 
          snew(gbin[i][j],nbins[2]); 
          snew(nmbin[i][j],nbins[2]); 
          snew(ngbin[i][j],nbins[2]); 
          for (k = 0 ; k<nbins[2] ; k++) {
            cbin[i][j][k]=0;
            mbin[i][j][k]=0.0;
            gbin[i][j][k]=0.0;
          }
        }
      }
      //determine offsets
      snew(offsets,3);
      for (i = 0 ; i<3 ; i++) {
        snew(offsets[i],3);
      }
      for (i=0 ; i<3 ; i++) {
        for (j=0 ; j<i ; j++) {
          (offsets)[i][j] = (int)(fr.box[i][j]/bindim);
        }
      }
     
      fprintf(stderr,"for cbins - loc_curv_bin_length:%d\n",loc_curv_bin_length);
     
      for (i=0; i<loc_curv_bin_length; i++) {
        for (k = 0 ; k < DIM ; k++) {
          tmpvec[k]=(int)(loc_curv_bin[i][k]/fr.box[k][k]*nbins[k]);
          if (tmpvec[k]>=nbins[k]) { //MF SAFETY MEASURE
            fprintf(stderr,"tmp: %d, box: %d\nreducing tmp!\n",tmpvec[k],nbins[k]);
            tmpvec[k]=nbins[k]-1;
          }
        }
        cbin[tmpvec[0]][tmpvec[1]][tmpvec[2]]+=1;
        mbin[tmpvec[0]][tmpvec[1]][tmpvec[2]]+=loc_curv_bin[i][3];
        gbin[tmpvec[0]][tmpvec[1]][tmpvec[2]]+=loc_curv_bin[i][4];
      }
      for (i = 0 ; i < nbins[0] ; i++) {
        for (j = 0 ; j < nbins[1] ; j++) {
          for (k = 0 ; k < nbins[2] ; k++) {
            if (cbin[i][j][k]>=thresh2) { //this threshold is used several times - keep it consistent (> vs >=!)
              mbin[i][j][k]=mbin[i][j][k]/cbin[i][j][k]; //also in loc_color_image
              gbin[i][j][k]=gbin[i][j][k]/cbin[i][j][k];
              //FOR TEMP UNAVERAGED PICS (see below)####################
              if (mbin[i][j][k]>max_m) max_m=mbin[i][j][k];
              if (mbin[i][j][k]<min_m) min_m=mbin[i][j][k];
              if (gbin[i][j][k]>max_g) max_g=gbin[i][j][k];
              if (gbin[i][j][k]<min_g) min_g=gbin[i][j][k];
              //########################################################
            }
            //printf("cbin: %d mbin:%f %d/%d/%d\n",cbin[i][j][k],mbin[i][j][k],i,j,k); //MF DEBUG
          }
        }
      }
      //TEMPORARILY ALSO PRINT UNAVERAGED PIC AFTER ROTATIONS################################################
      if (min_g==0.0 && max_g==0) { //no div by 0...
        scale = 1.0;
      } else {
        printf("min_g %18.17f  max_g %18.17f\n",min_g,max_g);//MF DEBUG
        if (fabs(min_g)>fabs(max_g)) {
          scale = fabs(1.0/(2*min_g));
          min_g = - fabs(min_g);
        } else {
          scale = fabs(1.0/(2*max_g));
          min_g = - fabs(max_g);
        }
      }
      printf("min_g %18.17f  max_g %18.17f scale %18.17f\n",min_g,max_g,scale);//MF DEBUG
      char* gnua = "UA-loc-G.py";
      FILE* gpua;
      gpua = fopen(gnua,"w");
      loc_color_image(gpua, cbin, nbins, bindim, gbin, scale,  min_g, thresh2);
      fclose(gpua);

      if (min_m==0.0 && max_m==0) { //no div by 0...
        scale = 1.0;
      } else {
        printf("min_m %18.17f  max_m %18.17f\n",min_m,max_m);//MF DEBUG
        if (fabs(min_m)>fabs(max_m)) {
          scale = fabs(1.0/(2*min_m));
          min_m = - fabs(min_m);
        } else {
          scale = fabs(1.0/(2*max_m));
          min_m = - fabs(max_m);
        }
      }
      printf("min_m %18.17f  max_m %18.17f scale %18.17f\n",min_m,max_m,scale);//MF DEBUG
      char* mnua = "UA-loc-M.py";
      FILE* mpua;
      mpua = fopen(mnua,"w");
      loc_color_image(mpua, cbin, nbins, bindim, mbin, scale,  min_m, thresh2);
      fclose(mpua);
      min_g=0.0;max_g=0.0;min_m=0.0;max_m=0.0;
      //#####################################################################################################
     
#define mPBC(s,t,u) rpbc_array(mbin,s,t,u,nbins,offsets)
#define gPBC(s,t,u) rpbc_array(gbin,s,t,u,nbins,offsets)
#define cPBC(s,t,u) pbc_array(cbin,s,t,u,nbins,offsets)
      for (i = 0 ; i < nbins[0] ; i++) {
        for (j = 0 ; j < nbins[1] ; j++) {
          for (k = 0 ; k < nbins[2] ; k++) {
            if (cbin[i][j][k]>=thresh2) {
              counter=0;
              for (s=-nlc2 ; s<=nlc2 ; s++) {
                for (t=-nlc2 ; t<=nlc2 ; t++) {
                  for (u=-nlc2 ; u<=nlc2 ; u++) {
                    if (/*cbin[i+s][j+t][k+u]*/cPBC(i+s,j+t,k+u)>=thresh2) {
                      counter++;
                    }
                  }
                }
              }
              mbin[i][j][k]/=counter; // divide by number of cells the curv is distributed over
              gbin[i][j][k]/=counter;
            }
          }
        }
      }
      for (i = 0 ; i < nbins[0] ; i++) { //DOES THIS HAVE TO BE A SEPARATE LOOP? SPEED-UP POTENTIAL!!
        for (j = 0 ; j < nbins[1] ; j++) {
          for (k = 0 ; k < nbins[2] ; k++) {
            if (cbin[i][j][k]>=thresh2) {
              for (s=-nlc2 ; s<=nlc2 ; s++) {
                for (t=-nlc2 ; t<=nlc2 ; t++) {
                  for (u=-nlc2 ; u<=nlc2 ; u++) {
                    if (/*cbin[i+s][j+t][k+u]*/cPBC(i+s,j+t,k+u)>=thresh2) {
                      nmbin[i][j][k]+=mPBC(i+s,j+t,k+u);//mbin[i+s][j+t][k+u];
                      ngbin[i][j][k]+=gPBC(i+s,j+t,k+u);//gbin[i+s][j+t][k+u];
                    }
                  }
                }
              }
              if (nmbin[i][j][k]>max_m) max_m=nmbin[i][j][k];
              if (nmbin[i][j][k]<min_m) min_m=nmbin[i][j][k];
              if (ngbin[i][j][k]>max_g) max_g=ngbin[i][j][k];
              if (ngbin[i][j][k]<min_g) min_g=ngbin[i][j][k];
            }
          }
        }
      }
     
      if (min_g==0.0 && max_g==0) { //no div by 0...
        scale = 1.0;
      } else {
        printf("min_g %18.17f  max_g %18.17f\n",min_g,max_g);//MF DEBUG
        if (fabs(min_g)>fabs(max_g)) {
          scale = fabs(1.0/(2*min_g));
          min_g = - fabs(min_g);
        } else {
          scale = fabs(1.0/(2*max_g));
          min_g = - fabs(max_g);
        }
      }
      printf("min_g %18.17f  max_g %18.17f scale %18.17f\n",min_g,max_g,scale);//MF DEBUG
      char* gn = "loc-G.py";
      FILE* gp;
      gp = fopen(gn,"w");
      loc_color_image(gp, cbin, nbins, bindim, ngbin, scale,  min_g, thresh2);
      fclose(gp);
     
      if (min_m==0.0 && max_m==0) { //no div by 0...
        scale = 1.0;
      } else {
        printf("min_m %18.17f  max_m %18.17f\n",min_m,max_m);//MF DEBUG
        if (fabs(min_m)>fabs(max_m)) {
          scale = fabs(1.0/(2*min_m));
          min_m = - fabs(min_m);
        } else {
          scale = fabs(1.0/(2*max_m));
          min_m = - fabs(max_m);
        }
      }
      printf("min_m %18.17f  max_m %18.17f scale %18.17f\n",min_m,max_m,scale);//MF DEBUG
      char* mn = "loc-M.py";
      FILE* mp;
      mp = fopen(mn,"w");
      loc_color_image(mp, cbin, nbins, bindim, nmbin, scale,  min_m, thresh2);
      fclose(mp);
    }

    for (i=0;i<7;i++)
      global_curv[i]/=(xr+1)*(yr+1)*(zr+1);
    printf("GLOBAL Mean:%f GLOBAL Gauss:%f Area: %f\n",global_curv[0],global_curv[1],global_curv[2]);
    printf("alt GLOBAL Mean:%f GLOBAL Gauss:%f Another - G: %f M: %f\n",global_curv[3],global_curv[4],global_curv[5],global_curv[6]);
    
  } while(read_next_frame(status,&fr));

  sfree(global_curv);

  if (LV) {
    j=nbins[0]*nbins[1]*nbins[2]*(xr+1)*(yr+1)*(zr+1)+1;
    for (i=0;i<j;i++) {
      sfree(loc_curv_bin[i]);
    }
    sfree(loc_curv_bin);
    freebin(&mbin,nbins);
    freebin(&gbin,nbins);
    freebin(&cbin,nbins);
    freebin(&nmbin,nbins);
    freebin(&ngbin,nbins);
  }

  fclose(fp);
  sfree(n3210);
  thanx(stderr);
  
  return 0;
}
